# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Functions for computing eigenvectors/eigenvalues (e.g. Lanczos)."""
# pylint: disable=invalid-name
# Capital letters for matrices

import time
from absl import logging
import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.ops as ops
import jax.random as random
import jax.scipy.linalg
import numpy as np


def safe_eigh(a, UPLO=None, symmetrize_input=True):
  # TODO: Remove when issue with CUDA eigh is resolved
  eigs, eig_vecs = lax.cond(
      jnp.linalg.norm(a) > 0.0,
      a, lambda a: jax.scipy.linalg.eigh(a, UPLO, symmetrize_input), a,
      lambda tridiag: (jnp.zeros(a.shape[0]), jnp.eye(a.shape[0])))
  return eigs, eig_vecs


def lanczos_alg(matrix_vector_product,
                dim,
                order,
                rng_key,
                dynamic_unroll=True,
                use_jax=True,
                verbose=False):
  """Lanczos algorithm for tridiagonalizing a real symmetric matrix.

  This function applies Lanczos algorithm of a given order.  This function
  does full reorthogonalization.

  WARNING: This function may take a long time to jit compile (e.g. ~3min for
  order 90 and dim 1e7).

  Args:
    matrix_vector_product: Maps v -> Hv for a real symmetric matrix H.
      Input/Output must be of shape [dim].
    dim: Matrix H is [dim, dim].
    order: An integer corresponding to the number of Lanczos steps to take.
    rng_key: The jax PRNG key.
    dynamic_unroll: bool, False does static unroll (default), True uses
      jax.fori_loop for faster JIT compile times.
    use_jax: Whether or not to do the lanczos computation in jax or numpy.
      Can save memory to use numpy.
    verbose: Print intermediate computations stats.

  Returns:
    tridiag: A tridiagonal matrix of size (order, order).
    vecs: A numpy array of size (order, dim) corresponding to the Lanczos
      vectors.
  """
  if dynamic_unroll:
    assert use_jax, 'Dynamic unroll only available with JAX.'
    return _lanczos_alg_dynamic_unroll(
        matrix_vector_product, dim, order, rng_key)
  if use_jax:
    backend = jnp
    index_update = ops.index_update
  else:
    backend = np
    def _index_update(array, index, value):
      new_array = array.copy()
      new_array[index] = value
      return new_array
    index_update = _index_update
  tridiag = backend.zeros((order, order))
  vecs = backend.zeros((order, dim))

  init_vec = random.normal(rng_key, shape=(dim,))
  init_vec = init_vec / backend.linalg.norm(init_vec)
  vecs = index_update(vecs, 0, init_vec)

  beta = 0
  start = time.time()
  # TODO: Better to use lax.fori loop for faster compile?
  for i in range(order):
    if verbose:
      end = time.time()
      logging.info('Iter %d out of %d. Time: %f', i, order, end-start)
    v = vecs[i, :].reshape((dim,))
    if i == 0:
      v_old = 0
    else:
      v_old = vecs[i - 1, :].reshape((dim,))

    w = matrix_vector_product(v)
    assert (w.shape[0] == dim and len(w.shape) == 1), (
        'Output of matrix_vector_product(v) must be of shape [dim].')
    w = w - beta * v_old

    alpha = backend.dot(w, v)
    tridiag = index_update(tridiag, (i, i), alpha)
    w = w - alpha * v

    # Full Reorthogonalization. Vectorized implementation of Gram Schmidt
    coeffs = backend.dot(vecs, w)
    scaled_vecs = (vecs.transpose()*coeffs).transpose()
    scaled_vecs = backend.sum(scaled_vecs, axis=0)
    w -= scaled_vecs

    beta = backend.linalg.norm(w)

    if i + 1 < order:
      # Small beta (<1e-6) implies Lanczos has converged.
      # TODO: Refactor to not run the loop when beta < 1e-6
      beta_write = lax.cond(beta < 1e-6, beta, jnp.zeros_like, beta,
                            lambda x: x)
      w_write = lax.cond(beta < 1e-6, w, jnp.zeros_like, w / beta, lambda x: x)
      tridiag = index_update(tridiag, (i, i + 1), beta_write)
      tridiag = index_update(tridiag, (i + 1, i), beta_write)
      vecs = index_update(vecs, i + 1, w_write)

  return (tridiag, vecs)


def _lanczos_alg_dynamic_unroll(
    matrix_vector_product, dim, order, rng_key):
  """Lanczos with jax.fori_loop unroll - see docstring for lanczos_alg()."""
  backend = jnp
  index_update = ops.index_update
  tridiag = backend.zeros((order, order))
  vecs = backend.zeros((order, dim))

  init_vec = random.normal(rng_key, shape=(dim,))
  init_vec = init_vec / backend.linalg.norm(init_vec)
  vecs = index_update(vecs, 0, init_vec)

  beta = 0.

  def _body_fn_update_alpha(i, vecs, tridiag, beta):
    """Duplicated code from first half of body_fn() used for final iteration."""
    v = vecs[i, :].reshape((dim,))
    v_old = vecs[i - 1, :].reshape((dim,))
    w = matrix_vector_product(v)
    assert (w.shape[0] == dim and len(w.shape) == 1), (
        'Output of matrix_vector_product(v) must be of shape [dim].')
    w = w - beta * v_old
    alpha = backend.dot(w, v)
    tridiag = index_update(tridiag, (i, i), alpha)
    return tridiag

  def body_fn(i, vals):
    """Main body used for jax.fori_loop."""
    vecs, tridiag, beta = vals
    v = vecs[i, :].reshape((dim,))
    v_old = lax.cond(i == 0, None, lambda x: jnp.zeros(dim, jnp.float32),
                     vecs, lambda vecs: vecs[i - 1, :].reshape((dim,)))
    w = matrix_vector_product(v)
    assert (w.shape[0] == dim and len(w.shape) == 1), (
        'Output of matrix_vector_product(v) must be of shape [dim].')
    w = w - beta * v_old

    alpha = backend.dot(w, v)
    tridiag = index_update(tridiag, (i, i), alpha)
    w = w - alpha * v

    # Full Reorthogonalization. Vectorized implementation of Gram Schmidt
    coeffs = backend.dot(vecs, w)
    scaled_vecs = (vecs.transpose()*coeffs).transpose()
    scaled_vecs = backend.sum(scaled_vecs, axis=0)
    w -= scaled_vecs

    beta = backend.linalg.norm(w)

    # Small beta (<1e-6) implies Lanczos has converged.
    beta_write = lax.cond(beta < 1e-6, beta, jnp.zeros_like, beta, lambda x: x)
    w_write = lax.cond(beta < 1e-6, w, jnp.zeros_like, w / beta, lambda x: x)
    tridiag = index_update(tridiag, (i, i + 1), beta_write)
    tridiag = index_update(tridiag, (i + 1, i), beta_write)
    vecs = index_update(vecs, i + 1, w_write)
    return (lax.stop_gradient(vecs), lax.stop_gradient(tridiag),
            lax.stop_gradient(beta))

  vecs, tridiag, beta_final = jax.lax.fori_loop(
      0, order - 1, body_fn, (vecs, tridiag, beta))
  # Update tridiag one last time for final iteration
  tridiag = _body_fn_update_alpha(order - 1, vecs, tridiag, beta_final)
  return (tridiag, vecs)


def _make_pos(vecx):
  return jnp.maximum(vecx, 0)


############     Lanczos, Lagrangian, dual function     ############


def max_eigenvector_lanczos(matrix_vector_product, dim, order, key, scl=-1,
                            dynamic_unroll=True, use_safe_eig_vec=True):
  """Get (soft)max eigenvector via Lanczos + Scipy eigendecomp."""
  tridiag, vecs = lanczos_alg(matrix_vector_product, dim, order, key,
                              verbose=False, dynamic_unroll=dynamic_unroll)
  eigs_triag, eig_vecs = safe_eigh(tridiag)
  if scl < 0:
    # Get max eigenvector
    eig_vec = jnp.dot(jnp.transpose(vecs), eig_vecs[:, -1])
  else:
    # Softmax weighting of max eigenvector - better gradients?
    eig_softmax = jnp.exp(scl*eigs_triag -
                          jax.scipy.special.logsumexp(scl * eigs_triag))
    eig_vec = jnp.dot(jnp.transpose(vecs), jnp.dot(eig_vecs, eig_softmax))

  if use_safe_eig_vec:
    # To handle the case when the norm of the eigen vector is ~0.
    # This can happen when triag is rank deficient. To handle this corner case,
    # sample a new eigen-vector and remove components with respect to
    # all eigen-vecs with non-zero eigen vals to get a vector in the null-space
    # of the Hessian => Eigen vector correspoding to eig-val 0.
    # TODO: Possible suspect if Lanczos starts to diverge.
    def get_eig_vec(vals):
      key, (eig_vecs, vecs, eigs_triag) = vals
      eig_vec = jnp.dot(jnp.transpose(vecs), eig_vecs[:, -1])
      random_vec = jax.random.uniform(key, shape=eig_vec.shape)
      eig_vecs = jnp.dot(jnp.transpose(vecs), eig_vecs)
      coeffs = jnp.dot(random_vec, eig_vecs)
      scaled_vecs = (eig_vecs * coeffs * (eigs_triag > 1e-7))
      scaled_vecs = jnp.sum(scaled_vecs, axis=1)
      eig_vec = random_vec - scaled_vecs
      return eig_vec

    vals = key, (eig_vecs, vecs, eigs_triag)
    eig_norm = jnp.linalg.norm(eig_vec)
    eig_vec = lax.cond(eig_norm < 1e-7, vals, get_eig_vec, eig_vec, lambda x: x)
  eig_vec = eig_vec/jnp.linalg.norm(eig_vec)
  return jax.lax.stop_gradient(eig_vec)


def min_eigenvector_lanczos(matrix_vector_product, *args, **kwargs):
  # If for matrix H, eigenvector v has eigenvalue lambda, then for matrix -H,
  # v has eigenvalue -lambda. So we find max eigenvector for -H instead.
  neg_mat_vec_product = lambda v: -matrix_vector_product(v)
  return max_eigenvector_lanczos(neg_mat_vec_product, *args, **kwargs)


def max_eigenvector_exact(matrix_vector_product, vec_dim, scl=-1,
                          report_all=False):
  """Get max eigenvector via Scipy eigendecomp."""
  @jax.jit
  def batched_Hv(v_batched):
    return jax.vmap(matrix_vector_product)(v_batched)

  H = batched_Hv(jnp.eye(vec_dim))
  H = (H + H.T)/2
  eig_vals, eig_vecs = safe_eigh(H)
  if scl < 0:
    eig_vec = eig_vecs[:, -1]
  else:
    eig_softmax_weights = jnp.exp(
        scl*eig_vals - jax.scipy.special.logsumexp(scl*eig_vals))
    eig_vec = jnp.sum(eig_vecs * jnp.expand_dims(eig_softmax_weights, axis=0),
                      axis=1)
  eig_vec = jax.lax.stop_gradient(eig_vec)
  if report_all:
    return eig_vec, (eig_vals, eig_vecs, H)
  return eig_vec


def min_eigenvector_exact(matrix_vector_product, vec_dim, scl=-1,
                          report_all=False):
  neg_mat_vec_product = lambda v: -matrix_vector_product(v)
  return max_eigenvector_exact(neg_mat_vec_product, vec_dim,
                               scl=scl, report_all=report_all)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for cvxpy_verify.py."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import jax.numpy as jnp
import jax.random as random
from jax_verify.src.sdp_verify import utils
from jax_verify.tests.sdp_verify import test_utils


class ParamExtractionTest(parameterized.TestCase):
  """Test the functions extracting network parameters from functions."""

  def check_fun_extract(self, fun_to_extract, example_inputs):
    extracted_params = utils.get_layer_params(fun_to_extract, example_inputs)

    eval_original = fun_to_extract(example_inputs)
    eval_extracted = utils.predict_cnn(extracted_params, example_inputs)

    self.assertAlmostEqual(jnp.abs(eval_original - eval_extracted).max(), 0.0)

  def test_cnn_extract(self):
    """Test that weights from a CNN can be extracted."""
    key = random.PRNGKey(0)
    k1, k2 = random.split(key)

    input_sizes = (1, 2, 2, 1)
    layer_sizes = [input_sizes, {
        'n_h': 2,
        'n_w': 2,
        'n_cout': 2,
        'padding': 'VALID',
        'stride': 1,
        'n_cin': 1
    }, 3]
    cnn_params = test_utils.make_cnn_params(layer_sizes, k1)

    fun_to_extract = functools.partial(utils.predict_cnn, cnn_params)
    example_inputs = random.normal(k2, input_sizes)

    self.check_fun_extract(fun_to_extract, example_inputs)

  def test_cnn_withpreproc(self):
    """Test extraction of weights from a CNN with input preprocessing."""
    key = random.PRNGKey(0)
    k1, k2, k3, k4 = random.split(key, num=4)

    input_sizes = (1, 2, 2, 3)
    layer_sizes = [input_sizes, {
        'n_h': 2,
        'n_w': 2,
        'n_cout': 2,
        'padding': 'VALID',
        'stride': 1,
        'n_cin': 3
    }, 3]
    cnn_params = test_utils.make_cnn_params(layer_sizes, k1)
    example_inputs = random.normal(k2, input_sizes)
    input_mean = random.normal(k3, (3,))
    input_std = random.normal(k4, (3,))

    def fun_to_extract(inputs):
      inp = (inputs - input_mean) / input_std
      return utils.predict_cnn(cnn_params, inp)

    self.check_fun_extract(fun_to_extract, example_inputs)

  def test_mlp_extract(self):
    """Test that weights from a MLP can be extracted."""
    key = random.PRNGKey(0)
    k1, k2 = random.split(key)

    input_sizes = (5,)
    layer_sizes = (5, 8, 5)
    mlp_params = test_utils.make_mlp_params(layer_sizes, k1)

    fun_to_extract = functools.partial(utils.predict_mlp, mlp_params)
    example_inputs = random.normal(k2, input_sizes)
    self.check_fun_extract(fun_to_extract, example_inputs)

  def test_mlp_withpreproc(self):
    """Test extraction of weights from a MLP with input preprocessing."""
    key = random.PRNGKey(0)
    k1, k2, k3, k4 = random.split(key, num=4)

    input_sizes = (5,)
    layer_sizes = (5, 8, 5)
    mlp_params = test_utils.make_mlp_params(layer_sizes, k1)
    example_inputs = random.normal(k2, input_sizes)
    input_mean = random.normal(k3, input_sizes)
    input_std = random.normal(k4, input_sizes)

    def fun_to_extract(inputs):
      inp = (inputs - input_mean) / input_std
      return utils.predict_mlp(mlp_params, inp)

    self.check_fun_extract(fun_to_extract, example_inputs)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Fastlin."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import haiku as hk
import jax
import jax.numpy as jnp
import jax_verify


class FastLinBoundTest(parameterized.TestCase):

  def assertArrayAlmostEqual(self, lhs, rhs):
    diff = jnp.abs(lhs - rhs).max()
    self.assertAlmostEqual(diff, 0.)

  def test_fc_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def linear_model(inp):
      return hk.Linear(1)(inp)

    z = jnp.array([[1., 2., 3.]])
    params = {'linear':
              {'w': jnp.ones((3, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}
    input_bounds = jax_verify.IntervalBound(z-1., z+1.)
    fun = functools.partial(linear_model.apply, params)
    output_bounds = jax_verify.fastlin_bound_propagation(fun, input_bounds)

    self.assertTrue(jnp.all(output_bounds.lower_lin.lin_coeffs == 1.))
    self.assertTrue(jnp.all(output_bounds.lower_lin.offset == 2.))
    self.assertTrue(jnp.all(output_bounds.upper_lin.lin_coeffs == 1.))
    self.assertTrue(jnp.all(output_bounds.upper_lin.offset == 2.))
    self.assertArrayAlmostEqual(jnp.array([[0., 1., 2.]]),
                                output_bounds.reference.lower)
    self.assertArrayAlmostEqual(jnp.array([[2., 3., 4.]]),
                                output_bounds.reference.upper)

    self.assertAlmostEqual(5., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_conv2d_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def conv2d_model(inp):
      return hk.Conv2D(output_channels=1, kernel_shape=(2, 2),
                       padding='VALID', stride=1, with_bias=True)(inp)

    z = jnp.array([1., 2., 3., 4.])
    z = jnp.reshape(z, [1, 2, 2, 1])

    params = {'conv2_d':
              {'w': jnp.ones((2, 2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(conv2d_model.apply, params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.fastlin_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(8., output_bounds.lower)
    self.assertAlmostEqual(16., output_bounds.upper)

  def test_conv1d_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def conv1d_model(inp):
      return hk.Conv1D(output_channels=1, kernel_shape=2,
                       padding='VALID', stride=1, with_bias=True)(inp)

    z = jnp.array([3., 4.])
    z = jnp.reshape(z, [1, 2, 1])

    params = {'conv1_d':
              {'w': jnp.ones((2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(conv1d_model.apply, params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.fastlin_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(7., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_relu_fastlin(self):
    def relu_model(inp):
      return jax.nn.relu(inp)
    z = jnp.array([[-2., 3.]])

    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.fastlin_bound_propagation(relu_model,
                                                         input_bounds)

    self.assertArrayAlmostEqual(jnp.array([[0., 2.]]), output_bounds.lower)
    self.assertArrayAlmostEqual(jnp.array([[0., 4.]]), output_bounds.upper)


class IBPFastLinBoundTest(parameterized.TestCase):

  def assertArrayAlmostEqual(self, lhs, rhs):
    diff = jnp.abs(lhs - rhs).max()
    self.assertAlmostEqual(diff, 0.)

  def test_fc_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def linear_model(inp):
      return hk.Linear(1)(inp)

    z = jnp.array([[1., 2., 3.]])
    params = {'linear':
              {'w': jnp.ones((3, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}
    input_bounds = jax_verify.IntervalBound(z-1., z+1.)
    fun = functools.partial(linear_model.apply, params)
    output_bounds = jax_verify.ibpfastlin_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(5., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_conv2d_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def conv2d_model(inp):
      return hk.Conv2D(output_channels=1, kernel_shape=(2, 2),
                       padding='VALID', stride=1, with_bias=True)(inp)

    z = jnp.array([1., 2., 3., 4.])
    z = jnp.reshape(z, [1, 2, 2, 1])

    params = {'conv2_d':
              {'w': jnp.ones((2, 2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(conv2d_model.apply, params)
    input_bounds = jax_verify.IntervalBound(z-1., z+1.)
    output_bounds = jax_verify.ibpfastlin_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(8., output_bounds.lower)
    self.assertAlmostEqual(16., output_bounds.upper)

  def test_conv1d_fastlin(self):

    @hk.without_apply_rng
    @hk.transform
    def conv1d_model(inp):
      return hk.Conv1D(output_channels=1, kernel_shape=2,
                       padding='VALID', stride=1, with_bias=True)(inp)

    z = jnp.array([3., 4.])
    z = jnp.reshape(z, [1, 2, 1])

    params = {'conv1_d':
              {'w': jnp.ones((2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(conv1d_model.apply, params)
    input_bounds = jax_verify.IntervalBound(z-1., z+1.)
    output_bounds = jax_verify.ibpfastlin_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(7., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_relu_fastlin(self):
    def relu_model(inp):
      return jax.nn.relu(inp)
    z = jnp.array([[-2., 3.]])

    input_bounds = jax_verify.IntervalBound(z-1., z+1.)
    output_bounds = jax_verify.ibpfastlin_bound_propagation(relu_model,
                                                            input_bounds)

    self.assertArrayAlmostEqual(jnp.array([[0., 2.]]), output_bounds.lower)
    self.assertArrayAlmostEqual(jnp.array([[0., 4.]]), output_bounds.upper)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mechanism to combine input bounds from multiple methods."""

import jax
import jax.numpy as jnp
from jax_verify.src import bound_propagation


Tensor = jnp.ndarray


class IntersectionBound(bound_propagation.Bound):
  """Concretises to intersection of constituent bounds."""

  def __init__(self, *base_bounds: bound_propagation.Bound):
    self.base_bounds = base_bounds

  @property
  def lower(self) -> Tensor:
    return jnp.array([bound.lower for bound in self.base_bounds]).max(axis=0)

  @property
  def upper(self) -> Tensor:
    return jnp.array([bound.upper for bound in self.base_bounds]).min(axis=0)


class ConstrainedBound(bound_propagation.Bound):
  """Wraps a Bound with additional concrete constraints."""

  def __init__(
      self, base_bound: bound_propagation.Bound, lower: Tensor, upper: Tensor):
    self._base_bound = base_bound
    self._lower = lower
    self._upper = upper

  def unwrap(self) -> bound_propagation.Bound:
    return self._base_bound.unwrap()

  @property
  def lower(self) -> Tensor:
    return jnp.maximum(self._base_bound.lower, self._lower)

  @property
  def upper(self) -> Tensor:
    return jnp.minimum(self._base_bound.upper, self._upper)


class IntersectionBoundTransform(bound_propagation.BoundTransform):
  """Aggregates several bound transforms, intersecting their concrete bounds."""

  def __init__(self, *base_transforms: bound_propagation.BoundTransform):
    self._base_transforms = base_transforms

  def input_transform(
      self, index: int, lower_bound: Tensor, upper_bound: Tensor
      ) -> IntersectionBound:
    """Constructs initial input bounds for each constituent bound type.

    Args:
      index: Integer identifying the input node.
      lower_bound: Original concrete lower bound on the input.
      upper_bound: Original concrete upper bound on the input.

    Returns:
      Intersection of the constituent input bounds.
    """
    return IntersectionBound(*[
        transform.input_transform(index, lower_bound, upper_bound)
        for transform in self._base_transforms])

  def primitive_transform(
      self, index: int, primitive: jax.core.Primitive, *args, **kwargs
      ) -> IntersectionBound:
    """Propagates bounds for each constituent bound type.

    Args:
      index: Integer identifying the computation node.
      primitive: Primitive Jax operation to transform.
      *args: Arguments of the primitive, wrapped as `IntersectionBound`s.
      **kwargs: Keyword Arguments of the primitive.

    Returns:
      Intersection of the propagated constituent output bounds.
    """
    def base_args_for_arg(arg):
      if isinstance(arg, bound_propagation.Bound):
        return [ConstrainedBound(bound, arg.lower, arg.upper)
                for bound in arg.unwrap().base_bounds]
      else:
        # Broadcast over the intersection components.
        return [arg for _ in self._base_transforms]

    base_args = [base_args_for_arg(arg) for arg in args]
    return IntersectionBound(*[
        transform.primitive_transform(index, primitive, *args, **kwargs)
        for transform, *args in zip(
            self._base_transforms, *base_args)])



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Library functions for SDP verification of neural networks."""
# pylint: disable=invalid-name
# Capital letters for matrices

import collections
import functools

from absl import logging
import jax
import jax.numpy as jnp
import jax.scipy
from jax_verify.src.sdp_verify import eigenvector_utils
from jax_verify.src.sdp_verify import utils
import numpy as np
import optax
import tree

IntBound = utils.IntBound
boundprop = utils.boundprop
flatten = lambda x: utils.flatten(x, backend=jnp)


def dual_fun(verif_instance, dual_vars, key=None, n_iter=30, scl=-1,
             exact=False, dynamic_unroll=True, include_info=False):
  # pylint: disable=invalid-name
  """Returns the dual objective value.

  Args:
    verif_instance: a utils.SdpDualVerifInstance, the verification problem
    dual_vars: A list of dual variables at each layer
    key: PRNGKey passed to Lanczos
    n_iter: Number of Lanczos iterations to use
    scl: Inverse temperature in softmax over eigenvalues to smooth optimization
        problem (if negative treat as hardmax)
    exact: Whether to use exact eigendecomposition instead of Lanczos
    dynamic_unroll: bool. Whether to use jax.fori_loop for Lanczos for faster
      JIT compilation. Default is False.
    include_info: if True, also return an `info` dict of various other
      values computed for the objective

  Returns:
    Either a single float, the dual upper bound, or if ``include_info=True``,
    returns a pair, the dual bound and a dict containing debugging info
  """
  key = key if key is not None else jax.random.PRNGKey(0)
  assert isinstance(verif_instance, utils.SdpDualVerifInstance)
  bounds = verif_instance.bounds
  layer_sizes = utils.layer_sizes_from_bounds(bounds)
  layer_sizes_1d = [np.prod(np.array(i)) for i in layer_sizes]
  N = sum(layer_sizes_1d) + 1
  info = {}

  # Mean activations at each layer
  activations_center = [(b.lb + b.ub) / 2 for b in bounds]
  # Maximum deviation from mean activations
  radius = [(b.ub - b.lb) / 2 for b in bounds]

  inner_lagrangian = verif_instance.make_inner_lagrangian(dual_vars)
  lagrangian = _make_transformed_lagrangian(
      inner_lagrangian, activations_center, radius)

  # Construct c_lambda and g_lambda terms
  zeros = [jnp.zeros(sz) for sz in layer_sizes]
  c_lambda = lagrangian(zeros)
  g_lambda = jax.grad(lagrangian)(zeros)
  g_lambda = flatten(g_lambda)
  info['c_lambda'] = c_lambda

  def Hv(v):
    """Hessian-vector product for H_lambda - refer to docstring for `Av()`."""
    lag_grad = lambda v2: flatten(jax.grad(lagrangian)(v2))
    hv_v = jax.grad(lambda v2: jnp.vdot(lag_grad(v2), v))(zeros)
    hv_flat = flatten(hv_v)
    return hv_flat

  def Av(v):
    """Matrix-vector product.

    Args:
      v: vector, DeviceArray

    Returns:
      Av: vector, Device array. A is defined as diag(kappa) - M(lambda) where
          M(lambda) = [0, g_lambda';
                       g_lambda, H_lambda], and these terms correspond to
          L~(z) = c_lambda + g_lambda' z + z' H_lambda z
    """
    # Expand Mv=[0 g'; g H] [v0;v1] = [g'v1; v0*g + H(v1)] = [Mv0;Mv1]
    # Compute Mv0 term
    mv_zero = jnp.reshape(jnp.vdot(g_lambda, v[1:]), (1,))
    # Compute Mv1 term
    mv_rest = Hv(v[1:]) + v[0] * g_lambda
    mv = jnp.concatenate([mv_zero, mv_rest], axis=0)
    diag_kappa_v = jnp.reshape(dual_vars[-1], mv.shape) * v
    av = diag_kappa_v - mv
    return jnp.reshape(av, v.shape)

  # Construct dual function (dual_vars[-1]=kappa)
  if exact:
    eig_vec, eig_info = eigenvector_utils.min_eigenvector_exact(
        Av, N, scl=scl, report_all=True)
    info['eig_info'] = eig_info
  else:
    eig_vec = eigenvector_utils.min_eigenvector_lanczos(
        Av, N, min(N, n_iter), key, scl, dynamic_unroll=dynamic_unroll)
  info['eig_vec'] = eig_vec
  info['kappa'] = dual_vars[-1]
  hess_val = jnp.vdot(eig_vec, Av(eig_vec))/(jnp.vdot(eig_vec, eig_vec))
  hess_val = jnp.reshape(hess_val, ())

  # Form dual objective
  lambda_minus = jnp.minimum(hess_val, 0.)
  kappa_hat = jnp.maximum(0, dual_vars[-1] - lambda_minus)
  dual_val = c_lambda + 0.5 * jnp.sum(kappa_hat)
  if include_info:
    return dual_val, info
  return dual_val


def _make_transformed_lagrangian(lagrangian, activations_center, radius):
  """Returns a function that computes transformed Lagrangian L~(z).

  Args:
    lagrangian: function L(x), the lagrangian with fixed dual variables.
    activations_center: list of Device arrays corresponding to mean activations
      by layer.
    radius: list of Device arrays corresponding to the interval bound radii by
      layer.

  Returns:
    transformed_lagrangian: a function L~(z), defined as
      L~(z) = L(activations_center + z * radius).

  For compatibility with the paper, the inner optimization should have [-1, 1]
  element-wise constraints, so here we re-express:
    max_{x: a-rad<=x<=a+rad} L(x), as
    max_{z: -1<=z<=1} L~(z), with L~ defined as above.
  """
  def transformed_lagrangian(zs):
    zs = [a + z * r for (a, z, r) in zip(activations_center, zs, radius)]
    return lagrangian(zs)

  return transformed_lagrangian


def project_duals(dual_vars, dual_types):
  """Projects dual variables to satisfy dual constraints."""
  make_pos = lambda v: None if v is None else jnp.maximum(v, 0)
  _project = lambda v, t: make_pos(v) if t == DualVarTypes.INEQUALITY else v
  return jax.tree_multimap(_project, dual_vars, dual_types)


def solve_sdp_dual_simple(verif_instance, key=None, opt=None, num_steps=10000,
                          eval_every=1000, verbose=False,
                          use_exact_eig_eval=True, use_exact_eig_train=False,
                          n_iter_lanczos=100,
                          kappa_reg_weight=None, kappa_zero_after=None,
                          device_type=None):
  """Compute verified lower bound via dual of SDP relaxation.

  Args:
    verif_instance: a utils.SdpDualVerifInstance
    key: jax.random.PRNGKey, used for Lanczos
    opt: an optax.GradientTransformation instance, the optimizer.
      If None, defaults to Adam with learning rate 1e-3.
    num_steps: int, the number of outer loop optimization steps
    eval_every: int, frequency of running evaluation step
    verbose: bool, enables verbose logging
    use_exact_eig_eval: bool, whether to use exact eigendecomposition instead of
      Lanczos when computing evaluation loss
    use_exact_eig_train: bool, whether to use exact eigendecomposition instead
      of Lanczos during training
    n_iter_lanczos: int, number of Lanczos iterations
    kappa_reg_weight: float, adds a penalty of sum(abs(kappa_{1:N})) to loss,
      which regularizes kappa_{1:N} towards zero. Default None is disabled.
    kappa_zero_after: int, clamps kappa_{1:N} to zero after ``kappa_zero_after``
      steps. Default None is disabled.
    device_type: string, used to clamp to a particular hardware device. Default
      None uses JAX default device placement

  Returns:
    A pair. The first element is a float, the final dual loss, which forms a
    valid upper bound on the objective specified by ``verif_instance``. The
    second element is a dict containing various debug info.
  """
  assert device_type in (None, 'cpu', 'gpu'), 'invalid device_type'
  assert isinstance(verif_instance, utils.SdpDualVerifInstance), 'invalid type'

  key = key if key is not None else jax.random.PRNGKey(0)
  opt = opt if opt is not None else optax.adam(1e3)
  dual_vars = jax.tree_map(
      lambda s: None if s is None else jnp.zeros(s), verif_instance.dual_shapes)
  dual_vars = init_duals_ibp(verif_instance, dual_vars)

  # Define loss function
  def loss(dual_vars, exact=use_exact_eig_train):
    return _loss(dual_vars, exact)

  @functools.partial(jax.jit, static_argnums=(1,), backend=device_type)
  def _loss(dual_var, exact):
    loss_val, step_info = dual_fun(
        verif_instance, dual_var, key, n_iter=n_iter_lanczos, exact=exact,
        include_info=True)
    step_info['loss_val'] = loss_val
    return loss_val, step_info

  # Define a compiled update step
  grad = jax.jit(jax.grad(loss, has_aux=True), backend=device_type)

  @functools.partial(jax.jit, backend=device_type)
  def grad_step(params, opt_state):
    g, info = grad(params)
    updates, new_opt_state = opt.update(g, opt_state)
    new_params = optax.apply_updates(params, updates)
    return new_params, new_opt_state, info

  # Optimize parameters in a loop
  opt_state = opt.init(dual_vars)
  info = collections.defaultdict(list)
  loss_log = []
  best_loss = 1e9

  # Main loop
  for i in range(num_steps):
    dual_vars, opt_state, step_info = grad_step(dual_vars, opt_state)
    loss_val = step_info['loss_val']
    print(f'Iter {i}: Loss {loss_val}')
    best_loss = min(best_loss, loss_val)
    loss_log.append(loss_val)

    # Regularization of kappa
    if kappa_reg_weight is not None and kappa_reg_weight >= 0:
      onehot = jax.nn.one_hot([0], dual_vars[-1].shape[1])
      mask = jnp.ones_like(onehot) - onehot
      dual_vars[-1] -= mask * kappa_reg_weight
    if (kappa_zero_after is not None and kappa_zero_after >= 0 and
        i > kappa_zero_after):
      onehot = jax.nn.one_hot([0], dual_vars[-1].shape[1])
      dual_vars[-1] *= onehot

    dual_vars = project_duals(dual_vars, verif_instance.dual_types)

    if i % eval_every == 0:
      dual_val, _ = loss(dual_vars, exact=use_exact_eig_eval)
      info['steps'].append(i)
      info['loss_vals'].append(float(dual_val))
      if verbose:
        print(f'Dual iter {i}: Train loss: {loss_val} Loss {dual_val}')

  final_loss = float(loss(dual_vars, exact=use_exact_eig_eval)[0])
  info['final_dual_vars'] = dual_vars
  info['final_loss'] = final_loss
  info['loss_log'] = loss_log
  info['best_train_loss'] = best_loss
  return final_loss, info


def solve_sdp_dual(verif_instance, key=None, opt=None, num_steps=10000,
                   verbose=False, eval_every=1000, use_exact_eig_eval=True,
                   use_exact_eig_train=False, n_iter_lanczos=30, scl=-1.0,
                   lr_init=1e-3, steps_per_anneal=100, anneal_factor=1.0,
                   num_anneals=3, opt_name='adam', gd_momentum=0.9,
                   add_diagnostic_stats=False,
                   opt_multiplier_fn=None, init_dual_vars=None,
                   init_opt_state=None, opt_dual_vars=None,
                   kappa_reg_weight=None, kappa_zero_after=None,
                   device_type=None, save_best_k=1):
  # pylint: disable=g-doc-return-or-yield, g-doc-args
  """Compute verified lower bound via dual of SDP relaxation.

  NOTE: This method exposes many hyperparameter options, and the method
  signature is subject to change. We instead suggest using
  ``solve_sdp_dual_simple`` instead if you need a stable interface.
  """
  # NB: Whereas the rest of the code in this library is fairly top-down
  # readable, avoids excessive `if` statements, tries to make the code look
  # like the formalism, etc, this is not the case for this method.
  # This is essentially the outer loop, and includes all the debugging/logging/
  # optimization tricks we need to get/debug good results.
  #
  # NB: Time profiling: On toy VerifInstances, JIT compilation dominates time
  # cost: JIT compilation takes ~12s, then we do ~3000 steps/sec.
  assert device_type in (None, 'cpu', 'gpu'), 'invalid device_type'
  assert isinstance(verif_instance, utils.SdpDualVerifInstance), 'invalid type'
  key = key if key is not None else jax.random.PRNGKey(0)

  dual_vars = jax.tree_map(
      lambda s: None if s is None else jnp.zeros(s), verif_instance.dual_shapes)
  dual_vars = init_duals_ibp(verif_instance, dual_vars)

  if init_dual_vars is not None:
    # Casting, here for Colab. Essentially same as `dual_vars = init_dual_vars`
    dual_vars = utils.structure_like(init_dual_vars, dual_vars)
  if opt_dual_vars is not None:
    opt_dual_vars = utils.structure_like(opt_dual_vars, dual_vars)

  # Create optimizer
  if opt is None:
    if (isinstance(steps_per_anneal, float) or
        isinstance(steps_per_anneal, int)):
      anneal_steps = [steps_per_anneal*(i+1) for i in range(num_anneals)]
    else:
      anneal_steps = np.cumsum(steps_per_anneal)
    anneal_steps = jnp.array(anneal_steps)
    def lr_schedule(t):
      cur_epoch = jnp.minimum(num_anneals, jnp.sum(t > anneal_steps))
      return lr_init * jnp.float_power(anneal_factor, cur_epoch)
    opt_class = getattr(optax, opt_name)
    base_opt = (opt_class(1., momentum=gd_momentum) if opt_name == 'sgd' else
                opt_class(1.))
    opt = optax.chain(base_opt, optax.scale_by_schedule(lr_schedule))
    if opt_multiplier_fn:
      # NB: Interface very specific to tree.map_structure_with_path
      # Example: opt_multiplier_fn=lambda path: 0.1 if 'lam' in path else 1.0
      opt_multipliers = tree.map_structure_with_path(
          lambda path, v: opt_multiplier_fn(path), dual_vars)
      opt = optax.chain(base_opt, optax.scale_by_schedule(lr_schedule),
                        utils.scale_by_variable_opt(opt_multipliers))
    else:
      opt = optax.chain(base_opt, optax.scale_by_schedule(lr_schedule))

  # Define loss function
  def loss(dual_vars, loss_scl=scl, exact=use_exact_eig_train):
    return _loss(dual_vars, loss_scl, exact)

  @functools.partial(jax.jit, static_argnums=(1, 2), backend=device_type)
  def _loss(dual_var, loss_scl, exact):
    loss_val, step_info = dual_fun(
        verif_instance, dual_var, key, n_iter=n_iter_lanczos, exact=exact,
        scl=loss_scl, include_info=True)
    step_info['loss_val'] = loss_val
    return loss_val, step_info

  # Define a compiled update step
  grad = jax.jit(jax.grad(loss, has_aux=True), backend=device_type)

  @functools.partial(jax.jit, backend=device_type)
  def grad_step(params, opt_state):
    g, info = grad(params)
    updates, new_opt_state = opt.update(g, opt_state)
    new_params = optax.apply_updates(params, updates)
    info['g'] = g
    info['updates'] = updates
    return new_params, new_opt_state, info

  # Optimize parameters in a loop
  opt_state = opt.init(dual_vars)
  if init_opt_state:
    opt_state = utils.structure_like(init_opt_state, opt_state)
  info = collections.defaultdict(list)
  loss_log = []
  store_best = []
  recent_eig_vecs = collections.deque(maxlen=10)
  best_loss = 1e9
  last_H = None
  start_i = 0

  # Main loop
  for i in range(start_i, num_steps):
    dual_vars_prev = dual_vars
    dual_vars, opt_state, step_info = grad_step(dual_vars, opt_state)
    loss_val = step_info['loss_val']
    print(f'Iter {i}: Loss {loss_val}')
    best_loss = min(best_loss, loss_val)
    if add_diagnostic_stats:
      info['dual_vars'].append(dual_vars_prev)
      eig_vec = step_info['eig_vec']
      cosine_sims = []
      for prev_eig_vec in recent_eig_vecs:
        denom = jnp.sqrt(jnp.linalg.norm(eig_vec)*jnp.linalg.norm(prev_eig_vec))
        eig_sim = jnp.sum(prev_eig_vec * eig_vec) / denom
        cosine_sims.append(abs(float(eig_sim)))
      info['c_lambda'].append(float(step_info['c_lambda']))
      info['past_10_cosine_sims'].append(np.array(cosine_sims))
      info['g'].append(step_info['g'])
      info['updates'].append(step_info['updates'])
      if use_exact_eig_train:
        # The info is for -H, so to get smallest for H, take negative of max
        eig_vals = -step_info['eig_info'][0][-1:-20:-1]
        cur_H = step_info['eig_info'][2]
        diff_H = 0 if last_H is None else np.linalg.norm(cur_H - last_H)
        last_H = cur_H
        info['diff_H'].append(float(diff_H))
        info['smallest_20_eig_vals'].append(eig_vals)
      recent_eig_vecs.appendleft(eig_vec)

    loss_log.append(loss_val)
    if len(store_best) < save_best_k:
      store_best.append((loss_val, dual_vars_prev))
      store_best.sort(key=lambda x: x[0])
    elif loss_val < store_best[-1][0]:
      store_best[-1] = (loss_val, dual_vars_prev)
      store_best.sort(key=lambda x: x[0])

    # Regularization of kappa
    if kappa_reg_weight is not None and kappa_reg_weight >= 0:
      onehot = jax.nn.one_hot([0], dual_vars[-1].shape[1])
      mask = jnp.ones_like(onehot) - onehot
      dual_vars[-1] -= mask * kappa_reg_weight
    if (kappa_zero_after is not None and kappa_zero_after >= 0 and
        i > kappa_zero_after):
      onehot = jax.nn.one_hot([0], dual_vars[-1].shape[1])
      dual_vars[-1] *= onehot

    dual_vars = project_duals(dual_vars, verif_instance.dual_types)

    if opt_dual_vars:
      distance_to_opt = jax.tree_multimap(lambda x, y: jnp.linalg.norm(x - y),
                                          dual_vars, opt_dual_vars)
      info['distance_to_opt'].append(distance_to_opt)

    if i % eval_every == 0:
      dual_val, _ = loss(dual_vars, loss_scl=-1, exact=use_exact_eig_eval)
      info['steps'].append(i)
      info['loss_vals'].append(float(dual_val))
      if verbose:
        print(f'Dual iter {i}: Train loss: {loss_val} Loss {dual_val}')

  final_loss = float(loss(dual_vars, loss_scl=-1, exact=use_exact_eig_eval)[0])
  info['final_dual_vars'] = dual_vars
  info['final_opt_state'] = opt_state
  info['final_loss'] = final_loss
  info['loss_log'] = loss_log
  info['store_best'] = store_best

  return final_loss, info

solve_dual_sdp_elided = solve_sdp_dual  # Alias

############     Dual initialization     ############

DualVar = utils.DualVar
DualVarFin = utils.DualVarFin
DualVarTypes = utils.DualVarTypes


def init_duals(verif_instance, key):
  """Initialize dual variables to zeros."""
  del key  # unused
  assert isinstance(verif_instance, utils.SdpDualVerifInstance)
  dual_shapes, _ = utils.get_dual_shapes_and_types(verif_instance.bounds)
  zeros_or_none = lambda s: None if s is None else jnp.zeros(s)
  return jax.tree_map(zeros_or_none, dual_shapes)


def _get_g_lambda(verif_instance, dual_vars):
  """Helper method for IBP initialization."""
  # NB: This code is (intentionally) copy-pasted from `dual_fun`, in order to
  # keep that method more top-to-bottom readable
  bounds = verif_instance.bounds
  # Mean activations at each layer
  activations_center = [(b.lb + b.ub) / 2 for b in bounds]
  # Maximum deviation from mean activations
  radius = [(b.ub - b.lb) / 2 for b in bounds]

  inner_lagrangian = verif_instance.make_inner_lagrangian(dual_vars)
  lagrangian = _make_transformed_lagrangian(
      inner_lagrangian, activations_center, radius)

  g_lambda = jax.grad(lagrangian)(activations_center)
  g_lambda = flatten(g_lambda)
  return jnp.reshape(g_lambda, (1, -1))


def init_duals_ibp(verif_instance, dual_vars):
  """Closed-form solution for dual variables which recovers IBP bound."""
  zero_duals = jax.tree_map(lambda x: x * 0., dual_vars)
  g_lambda = _get_g_lambda(verif_instance, zero_duals)
  kappa_opt_zero = jnp.reshape(jnp.sum(jnp.abs(g_lambda)), (1, 1))
  kappa_opt = jnp.concatenate([kappa_opt_zero, jnp.abs(g_lambda)], axis=1)
  ibp_duals = zero_duals[:-1] + [kappa_opt]
  return ibp_duals



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Propagate the relaxation through the network.

This is accomplished by traversing the JaxPR representation of the computation
and translating the computational graph.
"""
import abc
import functools
from typing import Callable, Tuple, Dict, Union, List, Optional

import jax
from jax import lax
import jax.numpy as jnp
from jax_verify.src import bound_propagation
from jax_verify.src import ibp
import numpy as np

Tensor = jnp.ndarray


class RelaxVariable(bound_propagation.Bound):
  """Variable used to build relaxation."""

  def __init__(self, name, base_bound):
    self.shape = base_bound.lower.shape
    self.base_bound = base_bound
    self.name = name
    self.constraints = None

  def set_constraints(self, constraints):
    self.constraints = constraints

  @property
  def lower(self):
    return self.base_bound.lower

  @property
  def upper(self):
    return self.base_bound.upper


class LinearConstraint:
  """Linear constraint, to be encoded into a solver."""

  def __init__(self, vars_and_coeffs, bias, sense):
    self._vars_and_coeffs = vars_and_coeffs
    self._bias = bias
    self.sense = sense
    self.sample_dependent = bool(self._bias.shape)

  def bias(self, index: int):
    """Get the bias corresponding to the minibatch sample `index`.

    If bias has a dimension, it means it is sample dependent. Otherwise,
    the bias is the same for all samples.

    Args:
      index: Index in the batch for which to build the variable.
    Returns:
      bias: value of the bias.
    """
    return self._bias[index] if self.sample_dependent else self._bias

  def vars_and_coeffs(self, index: int):
    """Get the variable and coefficients corresponding to the sample `index`.

    If coeffs has a dimension, the coefficients are sample dependent. Otherwise,
    the coefficients are the same for all samples.

    Args:
      index: Index in the batch for which to build the variable.
    Returns:
      vars_and_coeffs: vars_and_coeffs list where the coefficients are the one
        corresponding to the sample `index`
    """
    if self.sample_dependent:
      return [(var, (cpts, coeffs[index]))
              for (var, (cpts, coeffs)) in self._vars_and_coeffs]
    else:
      return self._vars_and_coeffs

  def encode_into_solver(self, solver: 'RelaxationSolver', index: int):
    """Encode the linear constraints into the provided solver.

    Args:
      solver: RelaxationSolver to create the linear constraint into.
      index: Index in the batch for which to build the variable.
    """
    solver.create_linear_solver_constraint(self, index)


class RelaxActivationConstraint:
  """Linear constraint involved in the relaxation of an activation."""

  def __init__(self, outvar, invar, scale, bias, sense):
    """Represents the constraint outvar =(>)(<) scale * invar + bias."""
    self.outvar = outvar
    self.invar = invar
    self.scale = scale
    self.bias = bias
    self.sense = sense

  def encode_into_solver(self, solver: 'RelaxationSolver', index: int):
    """Encode the linear constraints into the provided solver.

    Args:
      solver: RelaxationSolver to create the linear constraint into.
      index: Index in the batch for which to build the variable.
    """
    biases = np.reshape(self.bias[index, ...], [-1])
    slopes = np.reshape(self.scale[index, ...], [-1])
    for act_index, (slope, bias) in enumerate(zip(slopes, biases)):
      solver.create_activation_solver_constraint(
          self, act_index, slope.item(), bias.item())


class RelaxationSolver(metaclass=abc.ABCMeta):
  """Abstract solver for the relaxation."""

  @abc.abstractmethod
  def create_solver_variable(self, relax_var: RelaxVariable, index: int):
    """Create a new bound-constrained solver variable based on a RelaxVariable.

    Args:
      relax_var: Variable generated by the relaxation bound propagation.
      index: Index in the batch for which to build the variable.
    """

  @abc.abstractmethod
  def create_linear_solver_constraint(self, constraint: LinearConstraint,
                                      index: int):
    """Create a new solver linear constraint.

    Args:
      constraint: Constraint generated by the relaxation bound propagation.
      index: Index in the batch for which to build the variable.
    """

  @abc.abstractmethod
  def create_activation_solver_constraint(
      self, constraint: RelaxActivationConstraint, act_index: int,
      slope: float, bias: float):
    """Create the linear constraint involved in the activation relaxation.

    Args:
      constraint: Constraint generated by the relaxation bound propagation.
      act_index: Index of the activation to encode (in the variables involved
        in constraint)
      slope : Slope coefficients of the linear inequality.
      bias: Bias of the linear inequality
    """

  @abc.abstractmethod
  def minimize_objective(self,
                         var_name: int,
                         objective: Tensor,
                         objective_bias: float,
                         time_limit: Optional[int]) -> Tuple[float, bool]:
    """Minimize a linear function.

    Args:
      var_name: Index of the variable to define a linear function over the
        components.
      objective: Coefficients of the linear function.
      objective_bias: Bias of the linear function.
      time_limit: Maximum solve time in ms. Use None for unbounded.
    Returns:
      val: Value of the minimum.
      status: Status of the optimization function.
    """


def solve_relaxation(solver_ctor: Callable[[], RelaxationSolver],
                     objective: Tensor,
                     objective_bias: float,
                     variable_opt: RelaxVariable,
                     env: Dict[int, Union[RelaxVariable, Tensor]],
                     index: int,
                     time_limit: Optional[int] = None) -> Tuple[float, bool]:
  """Solves the relaxation using the provided LP solver.

  Args:
    solver_ctor: Constructor for the solver.
    objective: Objective to optimize, given as an array of coefficients to be
      applied to the variable to form a linear objective function
    objective_bias: Bias to add to objective
    variable_opt: RelaxVariable over which the linear function to optimize
      is defined.
    env: Environment created by applying boundprop with relaxation.py
    index: The index in the minibatch for which the LP should be solved
    time_limit: Time limit on solver. None if unbounded.
  Returns:
    opt: Value of the solution found.
    status: Whether the optimal solution has been achieved.
  """
  solver = solver_ctor()
  for key in env.keys():
    if isinstance(env[key], RelaxVariable):
      variable = env[key]
      # Create the variable in the solver.
      solver.create_solver_variable(variable, index)
      # Create the constraints in the solver.
      if variable.constraints:
        for constraint in variable.constraints:
          constraint.encode_into_solver(solver, index)
  return solver.minimize_objective(variable_opt.name,
                                   objective, objective_bias, time_limit)


def _get_linear(primitive, outval, *eqn_invars, **params):
  """Get linear expressions corresponding to an affine layer.

  Args:
    primitive: jax primitive
    outval: dummy tensor shaped according to a single example's outputs
    *eqn_invars: Arguments of the primitive, wrapped as RelaxVariables
    **params: Keyword Arguments of the primitive.

  Returns:
    For each output component, a pair `(bias, coefficients)`, where
    `coefficients` is a list of `(component, coefficient)` pairs.
  """
  def funx(x):
    if isinstance(x, RelaxVariable):
      return jnp.zeros_like(jnp.expand_dims(x.lower[0, ...], 0))
    else:
      return x
  def fungrad(i, args):
    return jnp.reshape(primitive.bind(*args, **params)[0, ...], [-1])[i]
  results = []
  # Loop over output dimensions one at a time to avoid creating a large
  # materialized tensor
  # TODO: Replace with something more efficient
  for i in range(outval.size):
    fung = functools.partial(fungrad, i)
    bias, current_grad = jax.value_and_grad(fung)(
        [funx(x) for x in eqn_invars])
    coefficients = []
    for res in current_grad:
      components = jnp.flatnonzero(res)
      coefficients.append((components, res.ravel()[components]))
    results.append((bias, coefficients))
  return results


def _get_relu_relax(lower, upper):
  """Upper chord of triangle relu relaxation."""
  on = lower >= 0.
  amb = jnp.logical_and(lower < 0., upper > 0.)
  slope = jnp.where(on, jnp.ones_like(lower), jnp.zeros_like(lower))
  slope += jnp.where(amb, upper/jnp.maximum(upper-lower, 1e-12),
                     jnp.zeros_like(lower))
  bias = jnp.where(amb, -lower * upper/jnp.maximum(upper-lower, 1e-12),
                   jnp.zeros_like(lower))
  return slope, bias


def _relax_input(
    index: int, in_bounds: bound_propagation.Bound,
    ) -> RelaxVariable:
  """Generates the initial inputs for the relaxation.

  Args:
    index: Integer identifying the input node.
    in_bounds: Concrete bounds on the input node.
  Returns:
    `RelaxVariable` for the initial inputs.
  """
  # Wrap initial bound as RelaxVariable bound.
  in_variable = RelaxVariable(index, in_bounds)
  return in_variable


_affine_primitives_list = [lax.broadcast_in_dim_p, lax.add_p,
                           lax.conv_general_dilated_p,
                           lax.dot_general_p, lax.sub_p,
                           lax.mul_p]
_activation_list = [lax.max_p]
_passthrough_list = [lax.reshape_p]


def _relax_primitive(
    index: int, out_bounds: bound_propagation.Bound,
    primitive: jax.core.Primitive, *args, **kwargs
    ) -> RelaxVariable:
  """Generates the relaxation for a given primitive op.

  Args:
    index: Integer identifying the computation node.
    out_bounds: Concrete bounds on the outputs of the primitive op.
    primitive: jax primitive
    *args: Arguments of the primitive, wrapped as RelaxVariables
    **kwargs: Keyword Arguments of the primitive.
  Returns:
    `RelaxVariable` that contains the output of this primitive for the
    relaxation, with all the constraints linking the output to inputs.
  """
  # Create variable for output of this primitive
  out_variable = RelaxVariable(index, out_bounds)
  # Create constraints linking output and input of primitive
  constraints = []
  if primitive in _passthrough_list:
    invar = args[0]
    constraints = [RelaxActivationConstraint(out_variable,
                                             invar,
                                             jnp.ones_like(invar.lower),
                                             jnp.zeros_like(invar.lower),
                                             0)]
  elif primitive in _affine_primitives_list:
    results = _get_linear(primitive, out_bounds.lower[0, ...],
                          *args, **kwargs)
    for i, (bias, coeffs) in enumerate(results):
      # Coefficients of the input variable(s).
      vars_and_coeffs = [
          (arg, coeff) for arg, coeff in zip(args, coeffs)
          if isinstance(arg, RelaxVariable)]
      # Equate with the output variable, by using a coefficient of -1.
      out_coeff = (np.array([i], dtype=np.int64), np.array([-1.]))
      vars_and_coeffs.append((out_variable, out_coeff))
      constraints.append(LinearConstraint(vars_and_coeffs, bias, 0))
  elif primitive in _activation_list:
    if len(args) == 2:
      # Generate relu relaxation
      # Relu is implemented as max(x, 0) and both are treated as arguments
      # We find the one that is a RelaxVariable (the other would just be
      # a JAX literal)
      if isinstance(args[0], RelaxVariable):
        invar = args[0]
        if jnp.amax(jnp.abs(args[1])) != 0.:
          raise NotImplementedError('Unsupported activation function')
      elif isinstance(args[1], RelaxVariable):
        invar = args[1]
        if jnp.amax(jnp.abs(args[0])) != 0.:
          raise NotImplementedError('Unsupported activation function')
      else:
        raise NotImplementedError('Activations with multiple arguments not'
                                  'supported.')
      slope, bias = _get_relu_relax(invar.lower, invar.upper)
      constraints += [
          RelaxActivationConstraint(out_variable,
                                    invar,
                                    jnp.zeros_like(invar.lower),
                                    jnp.zeros_like(invar.lower),
                                    1),  # relu(x) >= 0
          RelaxActivationConstraint(out_variable,
                                    invar,
                                    jnp.ones_like(invar.lower),
                                    jnp.zeros_like(invar.lower),
                                    1),  # relu(x) >= x
          RelaxActivationConstraint(out_variable,
                                    invar,
                                    slope,
                                    bias,
                                    -1)]  # upper chord of triangle relax
    else:
      raise NotImplementedError('Activations with multiple arguments not'
                                'supported.')
  out_variable.set_constraints(constraints)
  return out_variable


class RelaxationTransform(bound_propagation.GraphTransform[RelaxVariable]):
  """Transform to produce `RelaxVariable`s for each op."""

  def __init__(self, boundprop_transform: bound_propagation.BoundTransform):
    """Defines relaxation constraint propagation.

    Args:
      boundprop_transform: Basic Jax primitive ops' equivalents for
        the underlying bound propagation method.
    """
    self._boundprop_transform = boundprop_transform

  def input_transform(self, index, lower_bound, upper_bound):
    in_bounds = self._boundprop_transform.input_transform(
        index, lower_bound, upper_bound)
    return _relax_input(index, in_bounds)

  def primitive_transform(self, index, primitive, *args, **params):
    interval_args = [arg.base_bound if isinstance(arg, RelaxVariable) else arg
                     for arg in args]
    out_bounds = self._boundprop_transform.primitive_transform(
        index, primitive, *interval_args, **params)
    return _relax_primitive(index, out_bounds, primitive, *args, **params)


_PRIMITIVES_TO_OPTIMIZE_INPUT = (lax.max_p,)


class OptimizedRelaxationTransform(
    bound_propagation.GraphTransform[RelaxVariable]):
  """Wraps a RelaxVariable-producing BoundTransform to add optimization."""

  def __init__(
      self,
      transform: bound_propagation.GraphTransform[RelaxVariable],
      solver_ctor: Callable[[], RelaxationSolver]):
    self._transform = transform
    self.solver_ctor = solver_ctor
    self.solvers: List[RelaxationSolver] = []

  def tightened_variable_bounds(self, variable: RelaxVariable) -> RelaxVariable:
    """Compute tighter bounds based on the LP relaxation.

    Args:
      variable: Variable as created by the base boundprop transform. This is a
        RelaxVariable that has already been encoded into the solvers.
    Returns:
      tightened_variable: New variable, with the same name, referring to the
        same activation but whose bounds have been optimized by the LP solver.
    """
    lbs = []
    ubs = []
    for solver in self.solvers:
      nb_targets = np.prod(variable.shape[1:])
      sample_lbs = []
      sample_ubs = []
      for target_idx in range(nb_targets):
        objective = (jnp.arange(nb_targets) == target_idx).astype(jnp.float32)
        lb, optimal_lb = solver.minimize_objective(
            variable.name, objective, 0., 0)
        assert optimal_lb
        neg_ub, optimal_ub = solver.minimize_objective(
            variable.name, -objective, 0., 0)
        assert optimal_ub
        sample_lbs.append(lb)
        sample_ubs.append(-neg_ub)
      lbs.append(sample_lbs)
      ubs.append(sample_ubs)

    tightened_base_bound = ibp.IntervalBound(
        jnp.reshape(jnp.array(lbs), variable.shape),
        jnp.reshape(jnp.array(ubs), variable.shape))
    tightened_variable = RelaxVariable(variable.name, tightened_base_bound)
    tightened_variable.set_constraints(variable.constraints)
    # TODO Now that we have obtained tighter bounds, we could make the
    # decision to encode them into the LP solver, which might make the problems
    # easier to solve. This howevere would not change the strength of the
    # relaxation.
    return tightened_variable

  def input_transform(self, index: int,
                      lower_bound: Tensor,
                      upper_bound: Tensor) -> RelaxVariable:
    in_bounds = self._transform.input_transform(
        index, lower_bound, upper_bound)
    for minibatch_index in range(in_bounds.shape[0]):
      # Create one solver instance for each problem in the batch because they
      # will have different constraints.
      if minibatch_index >= len(self.solvers):
        self.solvers.append(self.solver_ctor())
      solver = self.solvers[minibatch_index]
      solver.create_solver_variable(in_bounds, minibatch_index)
    return in_bounds

  def primitive_transform(self,
                          index: int,
                          primitive: jax.core.Primitive,
                          *args: Union[RelaxVariable, Tensor],
                          **params) -> RelaxVariable:
    inp_args = args
    if primitive in _PRIMITIVES_TO_OPTIMIZE_INPUT:
      # This is an activation, so we want to have tight bounds for it.
      # We will take the inputs and tighten all the bounds.
      inp_args = [self.tightened_variable_bounds(inp)
                  if isinstance(inp, RelaxVariable) else inp
                  for inp in args]

    out_bounds = self._transform.primitive_transform(
        index, primitive, *inp_args, **params)
    for minibatch_index, solver in enumerate(self.solvers):
      # Encode the new variable and the associated constraints.
      solver.create_solver_variable(out_bounds, minibatch_index)
      if out_bounds.constraints:
        for constraint in out_bounds.constraints:
          constraint.encode_into_solver(solver, minibatch_index)
    return out_bounds



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
import inspect
import os
import sys

sys.path.insert(0, os.path.abspath('../'))

import jax_verify

# -- Project information -----------------------------------------------------

project = 'jax_verify'
copyright = '2020, DeepMind'  # pylint: disable=redefined-builtin
author = 'DeepMind'

# -- General configuration ---------------------------------------------------

master_doc = 'index'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'member-order': 'bysource',
    'special-members': True,
    'exclude-members': '__repr__, __str__, __weakref__',
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_theme_options = {
    # 'collapse_navigation': False,
    # 'sticky_navigation': False,
}

# -- Source code links -------------------------------------------------------


def linkcode_resolve(domain, info):
  """Resolve a GitHub URL corresponding to Python object."""
  if domain != 'py':
    return None

  try:
    mod = sys.modules[info['module']]
  except ImportError:
    return None

  obj = mod
  try:
    for attr in info['fullname'].split('.'):
      obj = getattr(obj, attr)
  except AttributeError:
    return None
  else:
    obj = inspect.unwrap(obj)

  try:
    filename = inspect.getsourcefile(obj)
  except TypeError:
    return None

  try:
    source, lineno = inspect.getsourcelines(obj)
  except OSError:
    return None

  # TODO: support tags after we release an initial version.
  return 'https://github.com/deepmind/jax_verify/blob/master/jax_verify/%s#L%d#L%d' % (
      os.path.relpath(filename, start=os.path.dirname(
          jax_verify.__file__)), lineno, lineno + len(source) - 1)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: disable=invalid-name
"""Small helper functions for testing."""

import jax
import jax.numpy as jnp
import jax.random as random
from jax_verify.src.sdp_verify import utils
import numpy as np

################## Toy Networks ######################


def _init_fc_layer(key, n_in, n_out):
  k1, k2 = random.split(key)
  W = random.normal(k1, (n_in, n_out))/ n_in
  b = random.normal(k2, (n_out,))/ n_in
  return W, b


def _init_conv_layer(key, n_h, n_w, n_cout, n_cin):
  """Random weights for a conv layer.

  Args:
    key: PRNG key
    n_h: Kernel Height
    n_w: Kernel Width
    n_cout: Out Channels
    n_cin: In Channels
  Returns:
    W: weights of the conv filters
    b: biases for the conv layer
  """
  k1, k2 = random.split(key)
  W = random.normal(k1, (n_h, n_w, n_cin, n_cout))
  b = random.normal(k2, (n_cout,))
  return W, b


def make_mlp_params(layer_sizes, key):
  sizes = layer_sizes
  keys = random.split(key, len(sizes))
  params = list(map(_init_fc_layer, keys[1:], sizes[:-1], sizes[1:]))
  return params


def make_cnn_params(layer_sizes, key):
  """Initialize a random network with conv layers, followed by linear layers.

  Args:
    layer_sizes: Layer-0 is the shape of the input in the NHWC format
    e.g. (1, 32, 32, 3). List of layer-wise hyperpameters. Dict of the following
    form for a conv-layer: {'n_h': 3, 'n_w': 3, 'n_cout': 32,
    'padding': 'VALID', 'stride': 1, 'n_cin': 3, input_shape:}. Input shape only
    if it's the first layer. For an FC layer, a single int corresponding to the
    number of op neurons.
    key: PRNG key

  Returns:
    W: weights of the conv filters
    b: biases for the conv layer
  """
  sizes = layer_sizes
  keys = random.split(key, len(sizes))
  params = []

  # NHWC, assert square
  assert len(layer_sizes[0]) == 4
  assert layer_sizes[0][1] == layer_sizes[0][2]
  conv_check = 0
  input_shape = layer_sizes[0][1]

  for counter, size in enumerate(layer_sizes[1:]):
    if isinstance(size, dict):
      size['input_shape'] = input_shape
      size['W'], size['b'] = _init_conv_layer(keys[counter], size['n_h'],
                                              size['n_w'], size['n_cout'],
                                              size['n_cin'])
      if size['padding'] == 'VALID':
        input_shape = int(np.ceil(input_shape - size['n_h'] +1)/size['stride'])
      else:
        input_shape = int(np.ceil(input_shape/size['stride']))
      size['output_shape'] = input_shape
      params.append(size)

    elif isinstance(size, int):
      # Check layer is FC
      if conv_check == 0:
        input_shape = input_shape * input_shape * layer_sizes[counter]['n_cout']
        conv_check = 1
      params.append(_init_fc_layer(keys[counter], input_shape, size))
      input_shape = size
    else:
      raise NotImplementedError('Unknown layer')
  return params


################## Toy Verification Instances ####################


def make_toy_verif_instance(seed=None, label=None, target_label=None, nn='mlp'):
  """Mainly used for unit testing."""
  key = jax.random.PRNGKey(0) if seed is None else jax.random.PRNGKey(seed)
  if nn == 'mlp':
    layer_sizes = '5, 5, 5'
    layer_sizes = np.fromstring(layer_sizes, dtype=int, sep=',')
    params = make_mlp_params(layer_sizes, key)
    inp_shape = (1, layer_sizes[0])
  else:
    if nn == 'cnn_simple':
      pad = 'VALID'
      # Input and filter size match -> filter is applied at just one location.

    else:
      pad = 'SAME'
      # Input is padded on right/bottom to form 3x3 input

    layer_sizes = [(1, 2, 2, 1), {
        'n_h': 2,
        'n_w': 2,
        'n_cout': 2,
        'padding': pad,
        'stride': 1,
        'n_cin': 1
    }, 3]
    inp_shape = layer_sizes[0]
    params = make_cnn_params(layer_sizes, key)

  bounds = utils.boundprop(
      params,
      utils.IntBound(lb=np.zeros(inp_shape),
                     ub=1*np.ones(inp_shape),
                     lb_pre=None,
                     ub_pre=None)
  )
  target_label = 1 if target_label is None else target_label
  label = 2 if label is None else label
  verif_instance = utils.make_nn_verif_instance(
      params,
      bounds,
      target_label=target_label,
      label=label,
      input_bounds=(0., 1.))
  return verif_instance


def make_mlp_layer_from_conv_layer(layer_params, input_bounds):
  """Convert Conv Layer into equivalent MLP layer."""
  assert isinstance(layer_params, dict)
  assert layer_params['padding'] == 'SAME'
  assert layer_params['stride'] == 1
  assert layer_params['n_cin'] == 1
  # only 'SAME' padding supported for now with stride (1,1)
  # to be used for unit-test support only
  # TODO: Add support for 'VALID'

  inp_shape = (layer_params['input_shape'], layer_params['input_shape'])
  w, b = layer_params['W'], layer_params['b']
  op_shape = int(np.ceil(inp_shape[0] / layer_params['stride']))
  pad_h = max((op_shape - 1) * layer_params['stride'] + layer_params['n_h'] -
              inp_shape[0], 0)
  pad_t = pad_h // 2
  pad_b = pad_h - pad_t
  pad_inp_shape = [inp_shape[0] + pad_h, inp_shape[1] + pad_h]
  padded_bounds = jnp.zeros(pad_inp_shape)
  lb = jax.ops.index_add(padded_bounds, jax.ops.index[pad_t:-pad_b,
                                                      pad_t:-pad_b],
                         input_bounds.lb[0, :, :, 0])
  ub = jax.ops.index_add(padded_bounds, jax.ops.index[pad_t:-pad_b,
                                                      pad_t:-pad_b],
                         input_bounds.ub[0, :, :, 0])
  pad_filter_shape = pad_inp_shape + [inp_shape[0], inp_shape[1], w.shape[-1]]
  pad_filter = jnp.zeros(pad_filter_shape)
  pad_bias = jnp.zeros(inp_shape + (w.shape[-1],))
  n_h, n_w = w.shape[0], w.shape[1]

  # unrolling the conv into an FC layer, stride=(1,1)
  for i in range(inp_shape[0]):
    for j in range(inp_shape[1]):
      pad_filter = jax.ops.index_add(
          pad_filter, jax.ops.index[i:i + n_h, j:j + n_w, i, j, 0], w[:, :, 0,
                                                                      0])
      pad_bias = jax.ops.index_add(pad_bias, jax.ops.index[i, j, 0], b[0])
      pad_filter = jax.ops.index_add(
          pad_filter, jax.ops.index[i:i + n_h, j:j + n_w, i, j, 1], w[:, :, 0,
                                                                      1])
      pad_bias = jax.ops.index_add(pad_bias, jax.ops.index[i, j, 1], b[1])
  pad_filter_lin = jnp.reshape(
      pad_filter,
      (pad_inp_shape[0] * pad_inp_shape[1], inp_shape[0] * inp_shape[1] * 2))
  pad_bias_lin = jnp.reshape(pad_bias, inp_shape[0] * inp_shape[1] * 2)

  return lb, ub, pad_filter_lin, pad_bias_lin


def make_mlp_verif_instance_from_cnn(verif_instance):
  """Convert CNN verif-instance into equivalent MLP verif-instance."""
  params_cnn = verif_instance.params_full
  assert not any([isinstance(x, dict) for x in params_cnn[1:]])
  # Only supports networks with structure conv-{fc}*
  weights = []
  for layer_params in params_cnn:
    if isinstance(layer_params, dict):
      lb, ub, w_lin, b_lin = make_mlp_layer_from_conv_layer(
          layer_params, verif_instance.bounds[0])
      lb = jnp.reshape(lb, (1, -1)) * 0.
      ub = jnp.reshape(ub, (1, -1)) * 1.
      weights.append((w_lin, b_lin))
    else:
      weights.append(layer_params)
  return lb, ub, weights



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run verification with out-of-the-box LP solver.

This example uses jax_verify to generate Linear Program (LP) constraints
expressed in CVXPY, which is then solved with a generic LP solver.

Note that this CVXPY example is purely illustrative - it incurs a large overhead
for defining the problem, since CVXPY struggles with the large number of
constraints, particularly with convolutional layers. We will release more
performant implementations with other LP solvers in the future. We also welcome
contributions.
"""
import functools
import pickle
from absl import app
from absl import flags
from absl import logging
import jax.numpy as jnp
import jax_verify
from jax_verify.src.sdp_verify import utils
import numpy as np

MLP_PATH = 'models/raghunathan18_pgdnn.pkl'
CNN_PATH = 'models/mnist_wongsmall_eps_10_adv.pkl'

flags.DEFINE_string('model', 'mlp', 'mlp or cnn or toy')
flags.DEFINE_string('boundprop_method', 'ibp', 'ibp or fastlin')
FLAGS = flags.FLAGS


def load_model(model_name):
  """Load model parameters and prediction function."""
  # Choose appropriate prediction function
  if model_name in ('mlp', 'toy'):
    model_path = MLP_PATH
    def model_fn(params, inputs):
      inputs = np.reshape(inputs, (inputs.shape[0], -1))
      return utils.predict_mlp(params, inputs)
  elif model_name == 'cnn':
    model_path = CNN_PATH
    model_fn = utils.predict_cnn
  else:
    raise ValueError('')

  # Get parameters
  if model_name == 'toy':
    params = [
        (np.random.normal(size=(784, 2)), np.random.normal(size=(2,))),
        (np.random.normal(size=(2, 10)), np.random.normal(size=(10,))),
    ]
  else:
    with jax_verify.open_file(model_path, 'rb') as f:
      params = pickle.load(f)
  return model_fn, params


def main(unused_args):

  # Load the parameters of an existing model.
  model_pred, params = load_model(FLAGS.model)
  logits_fn = functools.partial(model_pred, params)

  # Load some test samples
  with jax_verify.open_file('mnist/x_test_first100.npy', 'rb') as f:
    inputs = np.load(f)

  # Compute boundprop bounds
  eps = 0.1
  lower_bound = jnp.minimum(jnp.maximum(inputs[:2, ...] - eps, 0.0), 1.0)
  upper_bound = jnp.minimum(jnp.maximum(inputs[:2, ...] + eps, 0.0), 1.0)
  init_bound = jax_verify.IntervalBound(lower_bound, upper_bound)

  if FLAGS.boundprop_method == 'fastlin':
    final_bound = jax_verify.fastlin_bound_propagation(logits_fn, init_bound)
    boundprop_transform = jax_verify.fastlin_transform
  elif FLAGS.boundprop_method == 'ibp':
    final_bound = jax_verify.interval_bound_propagation(logits_fn, init_bound)
    boundprop_transform = jax_verify.ibp_transform
  else:
    raise NotImplementedError('Only ibp/fastlin boundprop are'
                              'currently supported')

  dummy_output = model_pred(params, inputs)

  # Run LP solver
  objective = jnp.where(jnp.arange(dummy_output[0, ...].size) == 0,
                        jnp.ones_like(dummy_output[0, ...]),
                        jnp.zeros_like(dummy_output[0, ...]))
  objective_bias = 0.
  value, status = jax_verify.solve_planet_relaxation(
      logits_fn, init_bound, boundprop_transform, objective,
      objective_bias, index=0)
  logging.info('Relaxation LB is : %f, Status is %s', value, status)
  value, status = jax_verify.solve_planet_relaxation(
      logits_fn, init_bound, boundprop_transform, -objective,
      objective_bias, index=0)
  logging.info('Relaxation UB is : %f, Status is %s', -value, status)

  logging.info('Boundprop LB is : %f', final_bound.lower[0, 0])
  logging.info('Boundprop UB is : %f', final_bound.upper[0, 0])


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bound propagation example usage: IBP, Fastlin, CROWN, CROWN-IBP.

Examples:
  python3 run_boundprop.py
  python3 run_boundprop.py --model=cnn
  python3 run_boundprop.py --boundprop_method=fastlin_bound_propagation
"""
import functools
import pickle
from absl import app
from absl import flags
from absl import logging
import jax.numpy as jnp
import jax_verify
from jax_verify.src.sdp_verify import utils
import numpy as np

MLP_PATH = 'models/raghunathan18_pgdnn.pkl'
CNN_PATH = 'models/mnist_wongsmall_eps_10_adv.pkl'
ALL_BOUNDPROP_METHODS = (
    jax_verify.interval_bound_propagation,
    jax_verify.fastlin_bound_propagation,
    jax_verify.ibpfastlin_bound_propagation,
    jax_verify.crown_bound_propagation,
    jax_verify.crownibp_bound_propagation,
)

flags.DEFINE_string('model', 'mlp', 'mlp or cnn')
flags.DEFINE_string('boundprop_method', '',
                    'Any boundprop method, such as `interval_bound_propagation`'
                    ' `fastlin_bound_propagation` `crown_bound_propagation`.'
                    'Empty string defaults to IBP.')
FLAGS = flags.FLAGS


def load_model(model_name):
  """Load model parameters and prediction function."""
  # Choose appropriate prediction function
  if model_name == 'mlp':
    model_path = MLP_PATH
    def model_fn(params, inputs):
      inputs = np.reshape(inputs, (inputs.shape[0], -1))
      return utils.predict_mlp(params, inputs)
  elif model_name == 'cnn':
    model_path = CNN_PATH
    model_fn = utils.predict_cnn
  else:
    raise ValueError('')

  # Load parameters from file
  with jax_verify.open_file(model_path, 'rb') as f:
    params = pickle.load(f)
  return model_fn, params


def main(unused_args):
  # Load some test samples
  with jax_verify.open_file('mnist/x_test_first100.npy', 'rb') as f:
    inputs = np.load(f)

  # Load the parameters of an existing model.
  model_pred, params = load_model(FLAGS.model)

  # Evaluation of the model on unperturbed images.
  clean_preds = model_pred(params, inputs)

  # Define initial bound
  eps = 0.1
  initial_bound = jax_verify.IntervalBound(
      jnp.minimum(jnp.maximum(inputs - eps, 0.0), 1.0),
      jnp.minimum(jnp.maximum(inputs + eps, 0.0), 1.0))

  # Because our function `model_pred` takes as inputs both the parameters
  # `params` and the `inputs`, we need to wrap it such that it only takes
  # `inputs` as parameters.
  logits_fn = functools.partial(model_pred, params)

  # Apply bound propagation. All boundprop methods take as an input the model
  # `function`, and the inital bounds, and return final bounds with the same
  # structure as the output of `function`. Internally, these methods work by
  # replacing each operation with its boundprop equivalent - see
  # bound_propagation.py for details.
  boundprop_method = (
      jax_verify.interval_bound_propagation if not FLAGS.boundprop_method else
      getattr(jax_verify, FLAGS.boundprop_method))
  assert boundprop_method in ALL_BOUNDPROP_METHODS, 'unsupported method'
  final_bound = boundprop_method(logits_fn, initial_bound)

  logging.info('Lower bound: %s', final_bound.lower)
  logging.info('Upper bound: %s', final_bound.upper)
  logging.info('Clean predictions: %s', clean_preds)

  assert jnp.all(final_bound.lower <= clean_preds), 'Invalid lower bounds'
  assert jnp.all(final_bound.upper >= clean_preds), 'Invalid upper bounds'


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for cvxpy_verify.py."""

import unittest
from absl.testing import absltest
from absl.testing import parameterized
from cvxpy.reductions.solvers.defines import INSTALLED_MI_SOLVERS as MIP_SOLVERS
import jax.numpy as jnp
from jax_verify.src.sdp_verify import cvxpy_verify
from jax_verify.src.sdp_verify import utils
from jax_verify.tests.sdp_verify import test_utils

NO_MIP_SOLVERS_MESSAGE = 'No mixed-integer solver is installed.'


class CvxpyTest(parameterized.TestCase):

  @unittest.skipUnless(MIP_SOLVERS, NO_MIP_SOLVERS_MESSAGE)
  def test_mip_status(self):
    """Test toy MIP is solved optimally by cvxpy."""
    for seed in range(10):
      verif_instance = test_utils.make_toy_verif_instance(seed)
      val, info = cvxpy_verify.solve_mip_mlp_elided(verif_instance)
      status = info['problem'].status
      assert val is not None
      assert status in ('optimal', 'optimal_inaccurate'), f'Status is {status}.'

  def test_sdp_status(self):
    """Test toy SDP is solved optimally by cvxpy."""
    for seed in range(10):
      verif_instance = test_utils.make_toy_verif_instance(seed)
      val, info = cvxpy_verify.solve_sdp_mlp_elided(verif_instance)
      status = info['problem'].status
      assert val is not None
      assert status in ('optimal', 'optimal_inaccurate'), f'Status is {status}.'


def _fgsm_example_and_bound(params, target_label, label):
  model_fn = lambda x: utils.predict_mlp(params, x)
  x = 0.5 * jnp.ones(utils.nn_layer_sizes(params)[0])
  epsilon = 0.5
  x_adv = utils.fgsm_single(model_fn, x, label, target_label, epsilon,
                            num_steps=30, step_size=0.03)
  return x_adv, utils.adv_objective(model_fn, x_adv, label, target_label)

MARGIN = 1e-6


class CrossingBoundsTest(parameterized.TestCase):
  """Check IBP,SDP relaxations <= MIP <= FGSM upper bound."""

  @unittest.skipUnless(MIP_SOLVERS, NO_MIP_SOLVERS_MESSAGE)
  def test_fgsm_vs_mip(self):
    num_repeats = 5
    target_label, label = 1, 2
    for seed in range(num_repeats):
      verif_instance = test_utils.make_toy_verif_instance(
          seed, target_label=target_label, label=label)
      mip_val, _ = cvxpy_verify.solve_mip_mlp_elided(verif_instance)
      _, fgsm_val = _fgsm_example_and_bound(
          verif_instance.params_full, target_label=target_label, label=label)
      assert mip_val > fgsm_val - MARGIN, (
          'MIP exact solution should be greater than FGSM lower bound.')

  @unittest.skipUnless(MIP_SOLVERS, NO_MIP_SOLVERS_MESSAGE)
  def test_sdp_vs_mip(self):
    num_repeats = 5
    loss_margin = 1e-3  # fixed via runs_per_test=300 with random seeds
    for seed in range(num_repeats):
      verif_instance = test_utils.make_toy_verif_instance(seed)
      mip_val, _ = cvxpy_verify.solve_mip_mlp_elided(verif_instance)
      sdp_val, _ = cvxpy_verify.solve_sdp_mlp_elided(verif_instance)
      assert sdp_val > mip_val - loss_margin, (
          'SDP relaxation should be greater than MIP exact solution. '
          f'Vals are MIP: {mip_val} SDP: {sdp_val}')


class MatchingBoundsTest(parameterized.TestCase):

  @unittest.skipUnless(MIP_SOLVERS, NO_MIP_SOLVERS_MESSAGE)
  def test_fgsm_vs_mip(self):
    """Check FGSM and MIP reach same solution/value most of the time."""
    # Note this test only works with fixed seeds
    num_repeats = 5
    expected_successes = 4
    num_successes = 0
    loss_margin = 0.01
    target_label, label = 1, 2
    for seed in range(num_repeats):
      verif_instance = test_utils.make_toy_verif_instance(
          seed, target_label=target_label, label=label)
      mip_val, _ = cvxpy_verify.solve_mip_mlp_elided(verif_instance)
      _, fgsm_val = _fgsm_example_and_bound(
          verif_instance.params_full, target_label=target_label, label=label)
      if abs(mip_val - fgsm_val) < loss_margin:
        num_successes += 1
    assert num_successes >= expected_successes, f'Successes: {num_successes}'


class SdpTest(parameterized.TestCase):

  def test_constraints_numpy(self):
    num_repeats = 5
    margin = 3e-4
    for seed in range(num_repeats):
      verif_instance = test_utils.make_toy_verif_instance(
          seed=seed, label=1, target_label=2)
      obj_value, info = cvxpy_verify.solve_sdp_mlp_elided(verif_instance)
      obj_np, violations = cvxpy_verify.check_sdp_bounds_numpy(
          info['P'].value, verif_instance)
      assert abs(obj_np - obj_value) < margin, 'objective does not match'
      for k, v in violations.items():
        assert v < margin, f'violation of {k} by {v}'


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for solving the convex relaxation using CVXPY."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import haiku as hk
import jax
import jax.numpy as jnp
import jax_verify
from jax_verify.src import bound_propagation
from jax_verify.src import cvxpy_relaxation_solver
from jax_verify.src import relaxation


class CVXPYRelaxationTest(parameterized.TestCase):

  def assertArrayAlmostEqual(self, lhs, rhs):
    diff = jnp.abs(lhs - rhs).max()
    self.assertAlmostEqual(diff, 0.)

  def get_bounds(self, fun, input_bounds):
    output = fun(input_bounds.lower)

    boundprop_transform = jax_verify.ibp_transform
    relaxation_transform = relaxation.RelaxationTransform(boundprop_transform)
    var, graph = bound_propagation.bound_propagation(
        relaxation_transform, fun, input_bounds)

    objective_bias = 0.
    index = 0

    lower_bounds = []
    upper_bounds = []
    for output_idx in range(output.size):
      objective = (jnp.arange(output.size) == output_idx).astype(jnp.float32)

      lower_bound, _ = relaxation.solve_relaxation(
          cvxpy_relaxation_solver.CvxpySolver, objective, objective_bias,
          var, graph.env, index)

      neg_upper_bound, _ = relaxation.solve_relaxation(
          cvxpy_relaxation_solver.CvxpySolver, -objective, objective_bias,
          var, graph.env, index)
      lower_bounds.append(lower_bound)
      upper_bounds.append(-neg_upper_bound)

    return jnp.array(lower_bounds), jnp.array(upper_bounds)

  def test_linear_cvxpy_relaxation(self):

    def linear_model(inp):
      return hk.Linear(1)(inp)

    z = jnp.array([[1., 2., 3.]])
    params = {'linear':
              {'w': jnp.ones((3, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(linear_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)

    lower_bounds, upper_bounds = self.get_bounds(fun, input_bounds)
    self.assertAlmostEqual(5., lower_bounds)
    self.assertAlmostEqual(11., upper_bounds)

  def test_conv1d_cvxpy_relaxation(self):

    def conv1d_model(inp):
      return hk.Conv1D(output_channels=1, kernel_shape=2,
                       padding='VALID', stride=1, with_bias=True)(inp)
    z = jnp.array([3., 4.])
    z = jnp.reshape(z, [1, 2, 1])

    params = {'conv1_d':
              {'w': jnp.ones((2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(conv1d_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)

    lower_bounds, upper_bounds = self.get_bounds(fun, input_bounds)

    self.assertAlmostEqual(7., lower_bounds)
    self.assertAlmostEqual(11., upper_bounds)

  def test_conv2d_cvxpy_relaxation(self):
    def conv2d_model(inp):
      return hk.Conv2D(output_channels=1, kernel_shape=(2, 2),
                       padding='VALID', stride=1, with_bias=True)(inp)
    z = jnp.array([1., 2., 3., 4.])
    z = jnp.reshape(z, [1, 2, 2, 1])

    params = {'conv2_d':
              {'w': jnp.ones((2, 2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(conv2d_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)

    lower_bounds, upper_bounds = self.get_bounds(fun, input_bounds)
    self.assertAlmostEqual(8., lower_bounds)
    self.assertAlmostEqual(16., upper_bounds)

  def test_relu_cvxpy_relaxation(self):
    def relu_model(inp):
      return jax.nn.relu(inp)
    z = jnp.array([[-2., 3.]])

    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    lower_bounds, upper_bounds = self.get_bounds(relu_model, input_bounds)

    self.assertArrayAlmostEqual(jnp.array([[0., 2.]]), lower_bounds)
    self.assertArrayAlmostEqual(jnp.array([[0., 4.]]), upper_bounds)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Propagate the bounds through the network.

This is accomplished by traversing the JaxPR representation of the computation
and translating the computational graph.
"""
import abc
import collections
from typing import Any, Callable, Dict, Generic, Sequence, Tuple, TypeVar, Union

import jax
import jax.numpy as jnp

from jax_verify.src import synthetic_primitives


Tensor = jnp.ndarray
T = TypeVar('T')
Nest = Union[T, Sequence[T], Dict[Any, T]]


class Bound(metaclass=abc.ABCMeta):
  """Abstract propagated bound."""

  @abc.abstractproperty
  def lower(self) -> Tensor:
    """Concrete lower bound."""

  @abc.abstractproperty
  def upper(self) -> Tensor:
    """Concrete upper bound."""

  def unwrap(self) -> 'Bound':
    """Underlying bound of method-specific type, without extra constraints.

    Usually this returns `self`. However, subclasses that wrap the bound to
    provide additional information (for example externally imposed interval
    constraints) will return the wrapped bound.

    Returns:
      Underlying bound arising directly from bound propagation.
    """
    return self


def unwrapping(fn):
  """Create a wrapper function to unwrap the bound arguments.

  Use as a decorator. If a propagation function has been defined assuming
  its input bound arguments are of the method-specific type
  (e.g. `fastlin.LinearBound`), then applying this decorator will allow the
  function to accept wrapped bound arguments with extra constraints
  (e.g. `intersection.ConstrainedBound`) which it will ignore.

  Args:
    fn: Function accepting possibly wrapped bound arguments.

  Returns:
    Function accepting bound arguments of the method-specific type.
  """
  unwrap = lambda x: x.unwrap() if isinstance(x, Bound) else x
  def fn_unwrapped(*args, **kwargs):
    return fn(*[unwrap(arg) for arg in args], **kwargs)
  return fn_unwrapped


Repr = TypeVar('Repr')


class GraphTransform(Generic[Repr], metaclass=abc.ABCMeta):
  """Abstract graph propagation method."""

  @abc.abstractmethod
  def input_transform(
      self,
      index: int,
      lower_bound: Tensor,
      upper_bound: Tensor,
      ) -> Repr:
    """Constructs input representations from lower/upper bound tensors.

    Args:
      index: Integer identifying the input node.
      lower_bound: Original concrete lower bound on the input.
      upper_bound: Original concrete upper bound on the input.

    Returns:
      Method-specific representation for the inputs.
    """

  @abc.abstractmethod
  def primitive_transform(
      self,
      index: int,
      primitive: jax.core.Primitive,
      *args: Union[Repr, Tensor],
      **params,
      ) -> Repr:
    """Applies the given primitive operation to its arguments' representations.

    Args:
      index: Integer identifying the computation node.
      primitive: Primitive Jax operation to transform.
      *args: Arguments of the primitive. Arguments are expressed as
        method-specific representations if they have any dependence on the
        network's original inputs, or tensors otherwise.
      **params: Keyword arguments of the primitive.

    Returns:
      Method-specific representation for the operation's output.
    """


BoundTransform = GraphTransform[Bound]


class OpwiseBoundTransform(BoundTransform):
  """Bound propagation method defined by functions for each primitive op."""

  def __init__(
      self,
      input_transform: Callable[[int, Tensor, Tensor], Bound],
      primitive_transform: Dict[jax.core.Primitive, Callable[..., Bound]]):
    self._input_transform = input_transform
    self._primitive_transform = primitive_transform

  def input_transform(self, index, lower_bound, upper_bound):
    return self._input_transform(index, lower_bound, upper_bound)

  def primitive_transform(self, index, primitive, *args, **params):
    if primitive not in self._primitive_transform:
      raise NotImplementedError(f'Unknown Primitive: {primitive}')
    return self._primitive_transform[primitive](index, *args, **params)


JaxVar = Union[jax.core.Var, jax.core.Literal]


class _Ref:

  def __init__(self, target_var: JaxVar):
    self.target_var: JaxVar = target_var


def _sub_graph(eqn):
  """Exposes the subgraph in a standard format if this is a jaxpr call.

  Args:
    eqn: A JaxPr equation that may contain a subgraph
  Returns:
    sub_graph: Can be None if there is no subgraph, or a JaxPR graph
  """
  if eqn.primitive == jax.custom_derivatives.custom_jvp_call_jaxpr_p:
    return eqn.params['fun_jaxpr'].jaxpr
  elif eqn.primitive == jax.xla.xla_call_p:
    return eqn.params['call_jaxpr']
  else:
    return None


class PropagationGraph(Generic[Repr]):
  """Holds a computational graph and the environment holding the variables.
  """

  def __init__(self, graph: jax.core.Jaxpr, graph_simplifier):
    self._graph = graph
    self.env: Dict[jax.core.Var, Union[Repr, Tensor, _Ref]] = {}
    self._reverse_env: Dict[jax.core.Var, Sequence[Repr]] = {}
    self.reverse_nodes: Dict[int, Repr] = {}
    self._first_index = None
    self._last_index = None
    self._graph_simplifier = graph_simplifier

  def read(self, var: JaxVar) -> Union[Repr, Tensor]:
    """Read the value from the environment."""
    if isinstance(var, jax.core.Literal):
      # Literals are values baked into the Jaxpr, e.g. ReLU's '0' arg to 'max'.
      return var.val
    else:
      val = self.env[var]
      if isinstance(val, _Ref):
        # The value is a reference to another variable. Follow the reference.
        return self.read(val.target_var)
      else:
        return val

  def build_inputs(
      self,
      transform: GraphTransform[Repr],
      bounds: Nest[Union[Bound, Tensor]],
      literals: Sequence[Tensor]):
    """Constructs initial representation of the graph's inputs.

    This is accomplished by inspecting the parameter inputs of the JaxPR
    representation of the computation, and constructing bounds using
    `input_transform`.

    Args:
      transform: Constructs initial input representation from
        lower/upper bound tensors.
      bounds: Nest of `IntervalBound` objects containing the lower and upper
        bounds on all the inputs, or `Tensor`s containing known inputs directly.
      literals: Values to use for any constant inputs of the JaxPR.
    """
    flat_bounds, _ = jax.tree_util.tree_flatten(bounds)
    invals = []
    index = 0
    for bound in flat_bounds:
      if isinstance(bound, Bound):
        input_val = transform.input_transform(index, bound.lower, bound.upper)
        index += 1
      else:
        # Input is a fixed tensor.
        input_val = bound
      invals.append(input_val)
    self.env.update(zip(self._graph.invars, invals))
    self.env.update(zip(self._graph.constvars, literals))

    self._first_index = index

  def forward_propagation(self, transform: GraphTransform[Repr]) -> Repr:
    """Performs forward propagation on the parsed computation graph.

    This is accomplished by traversing the JaxPR representation of the
    computation and translating the computational graph, using the
    `primitive_transform` replacement for each jax primitive.

    Args:
      transform: Basic Jax primitive ops' equivalents for operating on
        the representation (e.g. bound propagation).
    Returns:
      outvals: Propagated values corresponding to the graph output.
    """
    # Indices 0..k-1 are reserved for the k inputs to the network.
    index = self._first_index
    for eqn in self._graph_simplifier(self._graph).eqns:
      index = self._forward_prop_eqn(transform, index, eqn)
    self._last_index = index

    outvals = jax.tree_map(self.read, self._graph.outvars)
    return outvals

  def _forward_prop_eqn(self, transform, index, eqn):
    """Recursive step of `forward_propagation`."""
    sub_graph = _sub_graph(eqn)
    if sub_graph is not None:
      sub_graph = self._graph_simplifier(sub_graph)
      # Add the sub-graph's inputs to the environment.
      self.env.update({
          sub_invar: _Ref(invar)
          for sub_invar, invar in zip(sub_graph.invars, eqn.invars)})
      # Recursively propagate through the sub-graph.
      for sub_eqn in sub_graph.eqns:
        index = self._forward_prop_eqn(transform, index, sub_eqn)
      # Associate the sub-graph's outputs with the enclosing equation outputs.
      self.env.update({
          outvar: _Ref(sub_outvar)
          for sub_outvar, outvar in zip(sub_graph.outvars, eqn.outvars)})
      return index
    else:
      if len(eqn.outvars) != 1:
        # TODO Handle multi out primitives.
        raise NotImplementedError('Multiple outputs primitives not supported.')
      eqn_outvar = eqn.outvars[0]

      eqn_invars_vals = jax.tree_map(self.read, eqn.invars)
      eqn_params = eqn.params
      if any(isinstance(inval, Bound) for inval in eqn_invars_vals):
        eqn_outval = transform.primitive_transform(
            index, eqn.primitive, *eqn_invars_vals, **eqn_params)
        index += 1
      else:
        # No dependence on the network's inputs. Any primitive is supported.
        eqn_outval = eqn.primitive.bind(*eqn_invars_vals, **eqn_params)

      self.env[eqn_outvar] = eqn_outval
      return index

  def backward_propagation(self, primitive_transform, aggregate_function,
                           initial_outputs):
    """Performs backward propagation on the parsed computational graph.

    Args:
      primitive_transform: Dict matching basic jax primitives to functions
        (index, *input_bounds, **params) -> output_bounds, implementing
        the primitive ops' reverse bound propagation equivalents.
      aggregate_function: Function to aggregate the contribution of several
        of the network branches to one node. For example, for a ResNet, there
        will be several branches coming from the same node so when performing
        backward propagation, several contribution will come to the same node.
      initial_outputs: Values on the output to propagate backward.
    Returns:
      invals: Propagated values corresponding to the graph input.
    """

    flat_outs, _ = jax.tree_util.tree_flatten(initial_outputs)
    self._reverse_env = collections.defaultdict(list)
    for out_var, out_val in zip(self._graph.outvars, flat_outs):
      self._reverse_env[out_var].append(out_val)

    index = self._last_index
    for eqn in reversed(self._graph_simplifier(self._graph).eqns):
      index = self._backward_prop_eqn(primitive_transform, aggregate_function,
                                      index, eqn)

    if index != self._first_index:
      raise ValueError('Forward/backward indexing mismatch')

    invals = []
    for invar in reversed(self._graph.invars):
      if isinstance(self.read(invar), Bound):
        inval = aggregate_function(self._reverse_env.get(invar))
        index -= 1
        self.reverse_nodes[index] = inval
        invals.insert(0, inval)
      else:
        # Not a variable input.
        invals.insert(0, None)
    return invals

  def _backward_prop_eqn(self, primitive_transform, aggregate_function,
                         index, eqn):
    """Recursive step of `backward_propagation`."""
    sub_graph = _sub_graph(eqn)
    if sub_graph is not None:
      sub_graph = self._graph_simplifier(sub_graph)
      # Add the sub-graph's outputs to the environment.
      for sub_outvar, outvar in zip(sub_graph.outvars, eqn.outvars):
        self._reverse_env[sub_outvar].extend(self._reverse_env[outvar])
      # Recursively back-propagate through the sub-graph.
      for sub_eqn in reversed(sub_graph.eqns):
        index = self._backward_prop_eqn(primitive_transform, aggregate_function,
                                        index, sub_eqn)
      # Associate the sub-graph's inputs with the enclosing equation inputs.
      for sub_invar, invar in zip(sub_graph.invars, eqn.invars):
        self._reverse_env[invar].extend(self._reverse_env[sub_invar])
      return index

    else:
      if len(eqn.outvars) != 1:
        # TODO Handle multi out primitives.
        raise NotImplementedError('Multiple outputs primitives not supported.')

      if isinstance(self.read(eqn.outvars[0]), Bound):
        eqn_outval = aggregate_function(self._reverse_env[eqn.outvars[0]])
        eqn_invars_vals = jax.tree_map(self.read, eqn.invars)
        if eqn.primitive not in primitive_transform:
          raise NotImplementedError(f'Unknown Primitive: {eqn.primitive}')
        backward_prop_op = primitive_transform[eqn.primitive]
        index -= 1
        self.reverse_nodes[index] = eqn_outval
        eqn_invals = backward_prop_op(index, eqn_outval, *eqn_invars_vals,
                                      **eqn.params)
      else:
        # No dependence on the network's variable inputs.
        eqn_invals = [None for _ in eqn.invars]

      for in_var, in_val in zip(eqn.invars, eqn_invals):
        if not isinstance(in_var, jax.core.Literal):
          self._reverse_env[in_var].append(in_val)
      return index


def bound_propagation(
    transform: GraphTransform[Repr],
    function: Callable[..., Tensor],
    *bounds: Nest[Union[Bound, Tensor]],
    graph_simplifier=synthetic_primitives.activation_detector,
) -> Tuple[Repr, PropagationGraph[Repr]]:
  """Performs Bound Propagation on the model implemented by `function`.

  Args:
    transform: Basic Jax primitive ops' equivalents for operating on
      the representation.
    function: Pure function inputs -> outputs. If the function to propagate
      through has a more complex signature, the use of `functools.partial` can
      solve that problem.
    *bounds: Nest of `IntervalBound` objects containing the lower and upper
      bounds on all the inputs, or `Tensor`s containing known inputs directly.
    graph_simplifier: Function transforming the JaxPR graph into a simpler
      graph. Default value is a function identifying specific activation
      functions.
  Returns:
    bounds: Bounds over all the outputs of the function, with the same structure
      as the output of `function`
    graph: A PropagationGraph object which contains the parsed representation of
      the computation as well as the environment.
  """
  # Parse the computation graph.
  placeholder_inputs = jax.tree_util.tree_map(
      lambda b: b.lower if isinstance(b, Bound) else b,
      bounds)
  jaxpr_maker = jax.make_jaxpr(function)
  parsed = jaxpr_maker(*placeholder_inputs)
  output_shapes = jax.eval_shape(function, *placeholder_inputs)

  graph = PropagationGraph(parsed.jaxpr, graph_simplifier)
  graph.build_inputs(transform, bounds, parsed.literals)
  outvals = graph.forward_propagation(transform)

  # Make outvals into the same tree structure than the output of the function.
  tree_structure = jax.tree_util.tree_structure(output_shapes)
  outvals = jax.tree_util.tree_unflatten(tree_structure, outvals)

  return outvals, graph



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example architecture to test functions.
"""

import haiku as hk
import jax


class ResidualModel(hk.Module):
  """Model with several complexities.

  Contains Residual connections, stateful model (batch norms), nested modules.
  """

  def __init__(self):
    super().__init__()

    bn_config = {'create_scale': True,
                 'create_offset': True,
                 'decay_rate': 0.999}
   ## Definition of the modules.
    self.conv_block = hk.Sequential([
        hk.Conv2D(32, (3, 3), stride=3, rate=1), jax.nn.relu,
        hk.Conv2D(32, (3, 3), stride=3, rate=1), jax.nn.relu,
        hk.Conv2D(64, (3, 3), stride=3, rate=1), jax.nn.relu,
    ])

    self.conv_res_block = hk.Sequential([
        hk.Conv2D(32, (1, 1), stride=1, rate=1), jax.nn.relu,
        hk.Conv2D(32, (1, 1), stride=1, rate=1), jax.nn.relu,
        hk.Conv2D(64, (1, 1), stride=1, rate=1), jax.nn.relu,
    ])

    self.reshape_mod = hk.Flatten()

    self.lin_res_block = [
        (hk.Linear(128), hk.BatchNorm(name='lin_batchnorm_0', **bn_config)),
        (hk.Linear(256), hk.BatchNorm(name='lin_batchnorm_1', **bn_config))
    ]

    self.final_linear = hk.Linear(10)

  def call_all_act(self, inputs, is_training, test_local_stats=False):
    """Evaluate the model, returning its intermediate activations.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      all_acts: List with the intermediate activations of interest.
    """
    all_acts = []
    all_acts.append(inputs)

    ##  Forward propagation.
    # First conv layer.
    act = self.conv_block(inputs)
    all_acts.append(act)
    # Convolutional residual block.
    act = act + self.conv_res_block(act)
    all_acts.append(act)
    # Reshape before fully connected part.
    act = self.reshape_mod(act)
    all_acts.append(act)
    # Fully connected residual block.
    lin_block_act = act
    for lin_i, bn_i in self.lin_res_block:
      lin_block_act = lin_i(lin_block_act)
      lin_block_act = bn_i(lin_block_act, is_training, test_local_stats)
      lin_block_act = jax.nn.relu(lin_block_act)
    act = act + lin_block_act
    all_acts.append(act)
    # Final layer.
    act = self.final_linear(act)
    all_acts.append(act)
    return all_acts

  def __call__(self, inputs, is_training, test_local_stats=False):
    """Return only the final prediction of the model.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      pred: Array with the predictions, corresponding to the last activations.
    """
    all_acts = self.call_all_act(inputs, is_training, test_local_stats)
    return all_acts[-1]


class SmallResidualModel(hk.Module):
  """Small network with residual connections.

  Smaller version of ResidualModel.
  """

  def __init__(self):
    super().__init__()
    bn_config = {'create_scale': True,
                 'create_offset': True,
                 'decay_rate': 0.999}

    # Definition of the modules.
    self.conv_block = hk.Sequential([
        hk.Conv2D(1, (3, 3), stride=3, rate=1), jax.nn.relu,
        hk.Conv2D(1, (3, 3), stride=3, rate=1), jax.nn.relu,
    ])

    self.conv_res_block = hk.Sequential([
        hk.Conv2D(1, (1, 1), stride=1, rate=1), jax.nn.relu,
        hk.Conv2D(1, (1, 1), stride=1, rate=1), jax.nn.relu,
    ])

    self.reshape_mod = hk.Flatten()

    self.lin_res_block = [
        (hk.Linear(16), hk.BatchNorm(name='lin_batchnorm_0', **bn_config))
    ]

    self.final_linear = hk.Linear(10)

  def call_all_act(self, inputs, is_training, test_local_stats=False):
    """Evaluate the model, returning its intermediate activations.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      all_acts: List with the intermediate activations of interest.
    """
    all_acts = []
    all_acts.append(inputs)

    ##  Forward propagation.
    # First conv layer.
    act = self.conv_block(inputs)
    all_acts.append(act)
    # Convolutional residual block.
    act = act + self.conv_res_block(act)
    all_acts.append(act)
    # Reshape before fully connected part.
    act = self.reshape_mod(act)
    all_acts.append(act)
    # Fully connected residual block.
    lin_block_act = act
    for lin_i, bn_i in self.lin_res_block:
      lin_block_act = lin_i(lin_block_act)
      lin_block_act = bn_i(lin_block_act, is_training, test_local_stats)
      lin_block_act = jax.nn.relu(lin_block_act)
    act = act + lin_block_act
    all_acts.append(act)
    # Final layer.
    act = self.final_linear(act)
    all_acts.append(act)
    return all_acts

  def __call__(self, inputs, is_training, test_local_stats=False):
    """Return only the final prediction of the model.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      pred: Array with the predictions, corresponding to the last activations.
    """
    all_acts = self.call_all_act(inputs, is_training, test_local_stats)
    return all_acts[-1]


class TinyModel(hk.Module):
  """Tiny network.

  Single conv layer.
  """

  def __init__(self):
    super().__init__()
    # Definition of the modules.
    self.reshape_mod = hk.Flatten()

    self.lin_block = hk.Sequential([
        hk.Linear(20), jax.nn.relu,
    ])

    self.final_linear = hk.Linear(10)

  def call_all_act(self, inputs, is_training, test_local_stats=False):
    """Evaluate the model, returning its intermediate activations.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      all_acts: List with the intermediate activations of interest.
    """
    all_acts = []
    all_acts.append(inputs)
    act = inputs
    ##  Forward propagation.
    act = self.reshape_mod(act)
    all_acts.append(act)

    # First linear layer.
    act = self.lin_block(act)
    all_acts.append(act)
    # Final layer.
    act = self.final_linear(act)
    all_acts.append(act)
    return all_acts

  def __call__(self, inputs, is_training, test_local_stats=False):
    """Return only the final prediction of the model.

    Args:
      inputs: BHWC array of images.
      is_training: Boolean flag, whether this is during training.
      test_local_stats: Boolean flag, Whether local stats are used
        when is_training=False (for batchnorm).
    Returns:
      pred: Array with the predictions, corresponding to the last activations.
    """
    all_acts = self.call_all_act(inputs, is_training, test_local_stats)
    return all_acts[-1]




# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Interval Bound Propagation.
"""
from typing import Union, Tuple

import jax
from jax import lax
import jax.numpy as jnp

from jax_verify.src import bound_propagation
from jax_verify.src import synthetic_primitives
from jax_verify.src import utils

Tensor = jnp.ndarray
PrimitiveInput = Union[Tensor, bound_propagation.Bound]


class IntervalBound(bound_propagation.Bound):
  """Represent an interval where some activations might be valid."""

  def __init__(self, lower_bound: Tensor, upper_bound: Tensor):
    self.shape = lower_bound.shape
    self._lower_bound = lower_bound
    self._upper_bound = upper_bound

  @property
  def lower(self) -> Tensor:
    return self._lower_bound

  @property
  def upper(self) -> Tensor:
    return self._upper_bound

  def __add__(self, other: PrimitiveInput) -> 'IntervalBound':
    if isinstance(other, IntervalBound):
      return IntervalBound(self.lower + other.lower,
                           self.upper + other.upper)
    else:
      return IntervalBound(self.lower + other,
                           self.upper + other)

  def __radd__(self, other: PrimitiveInput) -> 'IntervalBound':
    return self.__add__(other)

  def __sub__(self, other: PrimitiveInput) -> 'IntervalBound':
    if isinstance(other, IntervalBound):
      return IntervalBound(self.lower - other.upper,
                           self.upper - other.lower)
    else:
      return IntervalBound(self.lower - other,
                           self.upper - other)

  def __rsub__(self, other: PrimitiveInput) -> 'IntervalBound':
    if isinstance(other, IntervalBound):
      return IntervalBound(other.lower - self.upper,
                           other.upper - self.lower)
    else:
      return IntervalBound(other - self.upper,
                           other - self.lower)


def _make_ibp_passthrough_primitive(primitive: jax.core.Primitive):
  """Generate a function that simply apply the primitive to the bounds.

  Args:
    primitive: jax primitive
  Returns:
    ibp_primitive: Function applying transparently the primitive to both
      upper and lower bounds.
  """
  def ibp_primitive(*args: PrimitiveInput, **kwargs) -> IntervalBound:
    # We assume that one version should be called with all the 'lower' bound and
    # one with the upper bound. If there is some argument that is not a bound,
    # we assumed it's just simple parameters and pass them through.
    lower_args = [arg.lower if isinstance(arg, bound_propagation.Bound) else arg
                  for arg in args]
    upper_args = [arg.upper if isinstance(arg, bound_propagation.Bound) else arg
                  for arg in args]

    out_lb = primitive.bind(*lower_args, **kwargs)
    out_ub = primitive.bind(*upper_args, **kwargs)
    return IntervalBound(out_lb, out_ub)
  return ibp_primitive


def _decompose_affine_argument(hs: PrimitiveInput) -> Tuple[Tensor, Tensor]:
  """Decompose an argument for a (bound, parameter) IBP propagation.

  We do not need to know which argument is the bound and which one is the
  parameter because we can simply decompose the operation.

  To propagate W * x, we can simply do
    out_diff = abs(W) * (x.upper - x.lower) / 2
    out_mean = W * (x.upper + x.lower) / 2
    out_ub = out_mean + out_diff
    out_lb = out_mean - out_diff

  This function will separate W and x in the two components we need so that we
  do not have to worry about argument order.

  Args:
    hs: Either an IntervalBound or a jnp.array
  Returns:
    part1, part2: jnp.arrays
  """
  if isinstance(hs, bound_propagation.Bound):
    return (hs.upper - hs.lower) / 2, (hs.upper + hs.lower) / 2
  else:
    return jnp.abs(hs), hs


def _ibp_conv_general_dilated(lhs: PrimitiveInput,
                              rhs: PrimitiveInput,
                              **kwargs
                              ) -> PrimitiveInput:
  """Propagation of IBP bounds through a convolution.

  Args:
    lhs: Input to the convolution.
    rhs: Kernel weights.
    **kwargs: Dict with the parameters of the convolution.
  Returns:
    out_bounds: IntervalBound on the output of the convolution.
  """
  req_arguments = ['window_strides', 'padding', 'lhs_dilation',
                   'rhs_dilation', 'dimension_numbers', 'feature_group_count',
                   'precision']
  lax_conv_params = utils.collect_required_arguments(req_arguments, kwargs)

  if (isinstance(lhs, bound_propagation.Bound) !=
      isinstance(rhs, bound_propagation.Bound)):
    lhses = _decompose_affine_argument(lhs)
    rhses = _decompose_affine_argument(rhs)

    forward_mean = lax.conv_general_dilated(lhses[1], rhses[1],
                                            **lax_conv_params)
    forward_range = lax.conv_general_dilated(lhses[0], rhses[0],
                                             **lax_conv_params)

    out_lb = forward_mean - forward_range
    out_ub = forward_mean + forward_range

    return IntervalBound(out_lb, out_ub)
  elif ((not isinstance(lhs, bound_propagation.Bound)) and
        (not isinstance(rhs, bound_propagation.Bound))):
    # Both are arrays, so can simply go through
    return lax.conv_general_dilated(lhs, rhs, **lax_conv_params)
  else:
    raise ValueError('BoundPropagation through general convolution '
                     'is not supported when both inputs are bounds.')


def _ibp_dot_general(lhs: PrimitiveInput,
                     rhs: PrimitiveInput,
                     **kwargs
                     ) -> PrimitiveInput:
  """Propagation of IBP bounds through a general dot product.

  We don't know if the bound is on the left or right hand side, but we expect
  that one hand is a bound and the other is a constant/parameter.

  Args:
    lhs: First input to the dot primitive.
    rhs: Second input to the dot primitive.
    **kwargs: Dict with the parameters of the general dot product.
  Returns:
    out_bounds: IntervalBound on the output of the dot product.
  """
  if (isinstance(lhs, bound_propagation.Bound) !=
      isinstance(rhs, bound_propagation.Bound)):
    lhses = _decompose_affine_argument(lhs)
    rhses = _decompose_affine_argument(rhs)

    forward_mean = lax.dot_general(lhses[1], rhses[1], **kwargs)
    forward_range = lax.dot_general(lhses[0], rhses[0], **kwargs)

    out_lb = forward_mean - forward_range
    out_ub = forward_mean + forward_range

    return IntervalBound(out_lb, out_ub)

  elif ((not isinstance(lhs, bound_propagation.Bound)) and
        (not isinstance(rhs, bound_propagation.Bound))):
    # Both are arrays, so can simply go through
    return lax.dot_general(lhs, rhs, **kwargs)
  else:
    raise ValueError('BoundPropagation through general dot product '
                     'is not supported when both inputs are bounds.')


def _ibp_mul(lhs: PrimitiveInput, rhs: PrimitiveInput) -> PrimitiveInput:
  """Propagation of IBP bounds through Elementwise multiplication.

  Args:
    lhs: Lefthand side of multiplication.
    rhs: Righthand side of multiplication.
  Returns:
    out_bounds: IntervalBound.
  """
  if (isinstance(lhs, bound_propagation.Bound) !=
      isinstance(rhs, bound_propagation.Bound)):
    # This is the case where one is a Bound and the other is not.
    lhses = _decompose_affine_argument(lhs)
    rhses = _decompose_affine_argument(rhs)

    forward_mean = lhses[1] * rhses[1]
    forward_range = lhses[0] * rhses[0]

    out_lb = forward_mean - forward_range
    out_ub = forward_mean + forward_range

    return IntervalBound(out_lb, out_ub)
  elif ((not isinstance(lhs, bound_propagation.Bound)) and
        (not isinstance(rhs, bound_propagation.Bound))):
    # Both are arrays, so can simply go through
    return lhs * rhs
  else:
    raise ValueError('BoundPropagation through multiply is not supported when '
                     'both inputs are bounds.')


def _ibp_div(lhs: PrimitiveInput, rhs: Tensor) -> PrimitiveInput:
  """Propagation of IBP bounds through Elementwise division.

  We don't support the propagation of bounds through the denominator.

  Args:
    lhs: Numerator of the division.
    rhs: Denominator of the division.
  Returns:
    out_bounds: Bound on the output of the division.
  """
  if isinstance(rhs, bound_propagation.Bound):
    raise ValueError('Bound propagation through the denominator unsupported.')
  return _ibp_mul(lhs, 1. / rhs)


@bound_propagation.unwrapping
def _ibp_add(lhs, rhs):
  """Propagation of IBP bounds through an addition.

  Args:
    lhs: Lefthand side of addition.
    rhs: Righthand side of addition.
  Returns:
    out_bounds: IntervalBound.
  """
  return lhs + rhs


@bound_propagation.unwrapping
def _ibp_sub(lhs, rhs):
  """Propagation of IBP bounds through a substraction.

  Args:
    lhs: Lefthand side of substraction.
    rhs: Righthand side of substraction.
  Returns:
    out_bounds: IntervalBound.
  """
  return lhs - rhs

# Define the mapping from jaxpr primitive to the IBP version.
_input_transform = lambda _, lower, upper: IntervalBound(lower, upper)
_primitives_to_passthrough = [lax.broadcast_in_dim_p,
                              lax.reduce_sum_p,
                              lax.max_p, lax.reshape_p,
                              synthetic_primitives.softplus_p]
_primitive_transform = {primitive: _make_ibp_passthrough_primitive(primitive)
                        for primitive in _primitives_to_passthrough}
_primitive_transform.update({
    lax.conv_general_dilated_p: _ibp_conv_general_dilated,
    lax.dot_general_p: _ibp_dot_general,
    lax.sub_p: _ibp_sub,
    lax.add_p: _ibp_add,
    lax.mul_p: _ibp_mul,
    lax.div_p: _ibp_div,
})
_primitive_transform = {key: utils.simple_propagation(prop_fun)
                        for key, prop_fun in _primitive_transform.items()}
bound_transform = bound_propagation.OpwiseBoundTransform(
    _input_transform, _primitive_transform)


def interval_bound_propagation(function, *bounds):
  """Performs IBP as described in https://arxiv.org/abs/1810.12715.

  Args:
    function: Function performing computation to obtain bounds for. Takes as
      only argument the network inputs.
    *bounds: jax_verify.IntervalBounds, bounds on the inputs of the function.
  Returns:
    output_bound: Bounds on the output of the function obtained by IBP
  """
  output_bound, _ = bound_propagation.bound_propagation(
      bound_transform, function, *bounds)
  return output_bound



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for propagating bounds through the networks defined in the Model Zoo.

We do not perform any check on the returned values but simply ensure that the
bound propagation can be performed on those networks.
"""
import pickle
from absl.testing import absltest
from absl.testing import parameterized

import haiku as hk
import jax
import jax.numpy as jnp
import jax_verify
from jax_verify.src import bound_propagation
from jax_verify.src import cvxpy_relaxation_solver
from jax_verify.src import relaxation
from jax_verify.tests import model_zoo
import numpy as np


class ModelZooModelTests(parameterized.TestCase):

  @parameterized.named_parameters(
      ('SmallResidualModel', model_zoo.SmallResidualModel),
      ('ResidualModel', model_zoo.ResidualModel),
      ('TinyModel', model_zoo.TinyModel)
  )
  def test_ibp(self, model_cls):

    @hk.transform_with_state
    def model_pred(inputs, is_training, test_local_stats=False):
      model = model_cls()
      return model(inputs, is_training, test_local_stats)

    inps = jnp.zeros((4, 28, 28, 1), dtype=jnp.float32)
    params, state = model_pred.init(jax.random.PRNGKey(42), inps,
                                    is_training=True)

    def logits_fun(inputs):
      return model_pred.apply(params, state, None, inputs,
                              False, test_local_stats=False)[0]

    input_bounds = jax_verify.IntervalBound(inps - 1.0, inps + 1.0)
    jax_verify.interval_bound_propagation(logits_fun, input_bounds)

  @parameterized.named_parameters(
      ('SmallResidualModel', model_zoo.SmallResidualModel),
      ('ResidualModel', model_zoo.ResidualModel),
      ('TinyModel', model_zoo.TinyModel)
  )
  def test_fastlin(self, model_cls):

    @hk.transform_with_state
    def model_pred(inputs, is_training, test_local_stats=False):
      model = model_cls()
      return model(inputs, is_training, test_local_stats)

    inps = jnp.zeros((4, 28, 28, 1), dtype=jnp.float32)
    params, state = model_pred.init(jax.random.PRNGKey(42), inps,
                                    is_training=True)

    def logits_fun(inputs):
      return model_pred.apply(params, state, None, inputs,
                              False, test_local_stats=False)[0]

    input_bounds = jax_verify.IntervalBound(inps - 1.0, inps + 1.0)
    jax_verify.fastlin_bound_propagation(logits_fun, input_bounds)

  @parameterized.named_parameters(
      ('SmallResidualModel', model_zoo.SmallResidualModel),
      ('ResidualModel', model_zoo.ResidualModel),
      ('TinyModel', model_zoo.TinyModel)
  )
  def test_ibpfastlin(self, model_cls):

    @hk.transform_with_state
    def model_pred(inputs, is_training, test_local_stats=False):
      model = model_cls()
      return model(inputs, is_training, test_local_stats)

    inps = jnp.zeros((4, 28, 28, 1), dtype=jnp.float32)
    params, state = model_pred.init(jax.random.PRNGKey(42), inps,
                                    is_training=True)

    def logits_fun(inputs):
      return model_pred.apply(params, state, None, inputs,
                              False, test_local_stats=False)[0]

    input_bounds = jax_verify.IntervalBound(inps - 1.0, inps + 1.0)
    jax_verify.ibpfastlin_bound_propagation(logits_fun, input_bounds)

  @parameterized.named_parameters(
      ('SmallResidualModel', model_zoo.SmallResidualModel),
      ('ResidualModel', model_zoo.ResidualModel),
      ('TinyModel', model_zoo.TinyModel)
  )
  def test_crownibp(self, model_cls):

    @hk.transform_with_state
    def model_pred(inputs, is_training, test_local_stats=False):
      model = model_cls()
      return model(inputs, is_training, test_local_stats)

    inps = jnp.zeros((4, 28, 28, 1), dtype=jnp.float32)
    params, state = model_pred.init(jax.random.PRNGKey(42), inps,
                                    is_training=True)

    def logits_fun(inputs):
      return model_pred.apply(params, state, None, inputs,
                              False, test_local_stats=False)[0]

    input_bounds = jax_verify.IntervalBound(inps - 1.0, inps + 1.0)
    jax_verify.crownibp_bound_propagation(logits_fun, input_bounds)

  @parameterized.named_parameters(
      ('SmallResidualModel', model_zoo.SmallResidualModel),
      ('TinyModel', model_zoo.TinyModel))
  def test_cvxpy_relaxation(self, model_cls):

    @hk.transform_with_state
    def model_pred(inputs, is_training, test_local_stats=False):
      model = model_cls()
      return model(inputs, is_training, test_local_stats)

    inps = jnp.zeros((4, 28, 28, 1), dtype=jnp.float32)
    params, state = model_pred.init(jax.random.PRNGKey(42), inps,
                                    is_training=True)

    def logits_fun(inputs):
      return model_pred.apply(params, state, None, inputs,
                              False, test_local_stats=False)[0]

    output = logits_fun(inps)
    input_bounds = jax_verify.IntervalBound(inps - 1.0, inps + 1.0)

    boundprop_transform = jax_verify.ibp_transform
    relaxation_transform = relaxation.RelaxationTransform(boundprop_transform)
    var, graph = bound_propagation.bound_propagation(
        relaxation_transform, logits_fun, input_bounds)

    objective_bias = 0.
    objective = jax.ops.index_update(jnp.zeros(output.shape[1:]), 0, 1)
    index = 0

    lower_bound, _ = relaxation.solve_relaxation(
        cvxpy_relaxation_solver.CvxpySolver, objective, objective_bias,
        var, graph.env, index)

    self.assertLessEqual(lower_bound, output[index, 0])


def _predict_mlp(params, inputs):
  # pylint: disable=invalid-name
  inputs = np.reshape(inputs, (inputs.shape[0], -1))
  for W, b in params[:-1]:
    outputs = jnp.dot(inputs, W) + b
    inputs = jnp.maximum(outputs, 0)
  W, b = params[-1]
  return jnp.dot(inputs, W) + b


class SavedModelTests(parameterized.TestCase):

  @parameterized.named_parameters(
      ('PGDNN', 'models/raghunathan18_pgdnn.pkl', 20, 19),
  )
  def test_mnist_mlp(self, model_name, num_examples, expected_correct):
    with jax_verify.open_file('mnist/x_test_first100.npy', 'rb') as f:
      mnist_x = np.load(f)
    with jax_verify.open_file('mnist/y_test.npy', 'rb') as f:
      mnist_y = np.load(f)
    with jax_verify.open_file(model_name, 'rb') as f:
      params = pickle.load(f)  # pytype: disable=wrong-arg-types  # due to GFile
    logits = np.array(_predict_mlp(params, mnist_x[:num_examples]))
    pred_labels = np.argmax(logits, axis=1)
    num_correct = np.sum(np.equal(mnist_y[:num_examples], pred_labels))
    print(num_correct)
    assert num_correct == expected_correct, f'Number correct: {num_correct}'


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Routines to solve the subproblem during verification."""
from typing import Dict, List, Tuple, Optional
from absl import logging
import cvxpy as cp

from jax_verify.src import relaxation
import numpy as np

_EPS = 1e-10
CvxpyConstraint = cp.constraints.constraint.Constraint
Tensor = np.ndarray


class CvxpySolver(relaxation.RelaxationSolver):
  """Holder class to represent problem being built."""

  def __init__(self):
    """Builds and solve the relaxation using CVXPY.
    """
    self.solver_variables: Dict[int, cp.Variable] = {}
    self.constraints: List[CvxpyConstraint] = []

  def create_solver_variable(self, relax_var: relaxation.RelaxVariable,
                             index: int):
    """Create a new bound-constrained CVXPY variable based on a RelaxVariable.

    Args:
      relax_var: Variable generated by the relaxation bound propagation.
      index: Index in the batch for which to build the variable.
    """
    lower = np.reshape(relax_var.lower[index, ...], [-1])
    upper = np.reshape(relax_var.upper[index, ...], [-1])
    var = cp.Variable(lower.shape)
    self.constraints += [lower <= var]
    self.constraints += [var <= upper]
    self.solver_variables[relax_var.name] = var

  def create_linear_solver_constraint(self,
                                      constraint: relaxation.LinearConstraint,
                                      index: int):
    """Create a new CVXPY linear constraint.

    Args:
      constraint: Constraint generated by the relaxation bound propagation.
      index: Index in the batch for which to build the variable.
    """
    rhs = cp.expressions.constants.Constant(constraint.bias(index))
    for (constraint_variable,
         (cpts, coeffs)) in constraint.vars_and_coeffs(index):
      if isinstance(constraint_variable, relaxation.RelaxVariable):
        current_variable = self.solver_variables[constraint_variable.name]
        for component, coeff in zip(cpts, coeffs):
          if np.abs(coeff) > _EPS:
            rhs += current_variable[component] * coeff.item()
    if constraint.sense == 0:
      self.constraints += [rhs == 0]
    if constraint.sense == 1:
      self.constraints += [rhs <= 0]
    if constraint.sense == -1:
      self.constraints += [rhs >= 0]

  def create_activation_solver_constraint(
      self, constraint: relaxation.RelaxActivationConstraint, act_index: int,
      slope: float, bias: float):
    """Create the linear constraint involved in the activation relaxation.

    Args:
      constraint: Constraint generated by the relaxation bound propagation.
      act_index: Index of the activation to encode (in the variables involved
        in constraint)
      slope : Slope coefficients of the linear inequality.
      bias: Bias of the linear inequality
    """
    outvar = self.solver_variables[constraint.outvar.name]
    invar = self.solver_variables[constraint.invar.name]
    expression = invar[act_index] * slope + bias - outvar[act_index]
    if constraint.sense == 0:
      self.constraints += [expression == 0]
    if constraint.sense == 1:
      self.constraints += [expression <= 0]
    if constraint.sense == -1:
      self.constraints += [expression >= 0]

  def minimize_objective(self, var_name: int,
                         objective: Tensor,
                         objective_bias: float,
                         time_limit: Optional[int] = None
                         ) -> Tuple[float, bool]:
    """Minimize a linear function.

    Args:
      var_name: Index of the variable to define a linear function over the
        components.
      objective: Coefficients of the linear function.
      objective_bias: Bias of the linear function.
      time_limit: Maximum solve time in ms. Use None for unbounded. CVXPY does
        not support time_limit so any other value will raise a ValueError.
    Returns:
      val: Value of the minimum.
      status: Status of the optimization function.
    """
    if time_limit:
      raise ValueError('Cvxpy Solver does not support time limit.')
    # Define the objective function
    obj = cp.expressions.constants.Constant(0.)
    for i, coeff in enumerate(objective):
      obj += self.solver_variables[var_name][i] * coeff.item()
    objective = cp.Minimize(obj)
    prob = cp.Problem(objective, self.constraints)
    logging.info('Starting problem solve.')
    prob.solve()
    logging.info('Problem status is %s', prob.status)
    val = obj.value
    if val is not None:
      val += objective_bias
    return val, (prob.status == cp.settings.OPTIMAL)
